from ..metamodel import ir, builtins, helpers, types
from ..metamodel.visitor import collect_by_type
from ..metamodel.util import FrozenOrderedSet, OrderedSet
from ..metamodel.compiler import group_tasks
from . import lqp_bundle
from . import utils, types as lqp_types
from .primitives import lqp_avg_op, lqp_operator, build_primitive
from .pragmas import pragma_to_lqp_name
from .types import meta_type_to_lqp
from .constructors import (
    mk_abstraction, mk_and, mk_exists, mk_or, mk_pragma, mk_primitive,
    mk_specialized_value, mk_type, mk_value, mk_attribute,
)
from .algorithms import (
    is_script, is_algorithm_script, is_algorithm_logical,
    is_while_loop, is_while_script, construct_monoid, is_empty_instruction
)
from .builtins import (
    has_global_annotation, get_upsert_annotation, get_monoid_annotation,
    get_monus_annotation, has_assign_annotation, get_arity, supported_lqp_annotations,
    is_lifted_data, has_gnf_data_annotation, get_gnf_column_name, ROW_ID_COLUMN,
)
from .utils import (
    TranslationCtx, ExportDescriptor, ExportInfo, gen_unique_var, gen_rel_id, extract_export_fqn
)
from .validators import assert_valid_input
from .rewrite.functional_dependencies import (
    normalized_fds, contains_only_declarable_constraints
)
from decimal import Decimal as PyDecimal
from datetime import datetime, date, timezone
from typing import Sequence, Tuple, cast, Union, Optional
from warnings import warn
import io
import pandas as pd
import re
import uuid
from collections import defaultdict

# Main access point for translating metamodel to lqp. Converts the model IR to an LQP epoch.
def to_lqp(model: ir.Model, fragment_name: bytes, ctx: TranslationCtx) -> tuple[Optional[ExportInfo], lqp_bundle.ir.Epoch]:
    assert_valid_input(model)
    decls: list[lqp_bundle.ir.Declaration] = []
    reads: list[lqp_bundle.ir.Read] = []

    # LQP only accepts logical tasks
    # These are asserted at init time
    root = cast(ir.Logical, model.root)
    for subtask in root.body:
        assert isinstance(subtask, ir.Logical)
        new_decls = _translate_to_decls(ctx, subtask)
        decls.extend(new_decls)

    # Add pyrel error attributes to reads.
    for err_id in _extract_pyrel_error_ids(ctx, model):
        ctx.output_ids.append(err_id)

    reads.extend(_get_output_reads(ctx.output_ids))

    # export_info is None if no exports, or a list of (fqn, filename, col_types) tuples
    export_info = None
    if len(ctx.export_descriptors) > 0:
        # Group descriptors by export FQN to create separate reads per export table
        exports_by_fqn = defaultdict(list)
        for descriptor in ctx.export_descriptors:
            exports_by_fqn[descriptor.export_fqn].append(descriptor)

        # Create a read for each export table and collect export info for each
        export_info = []
        for export_fqn, descriptors in exports_by_fqn.items():
            export_filename, col_types, export_reads = _get_export_reads(descriptors)
            reads.extend(export_reads)
            export_info.append((export_fqn, export_filename, col_types))

    debug_info = lqp_bundle.ir.DebugInfo(id_to_orig_name=ctx.rel_id_to_orig_name, meta=None)
    fragment_id = lqp_bundle.ir.FragmentId(id=fragment_name, meta=None)
    fragment = lqp_bundle.ir.Fragment(id=fragment_id, declarations=decls, meta=None, debug_info=debug_info)
    define_op = lqp_bundle.ir.Define(fragment=fragment, meta=None)

    txn = lqp_bundle.ir.Epoch(
        reads=reads,
        writes=[lqp_bundle.ir.Write(write_type=define_op, meta=None)],
        meta=None
    )

    return (export_info, txn)

def _effect_bindings(effect: Union[ir.Output, ir.Update]) -> list[ir.Value]:
    if isinstance(effect, ir.Output):
        # Unions may not return anything. The generated IR contains a None value when this
        # happens. We ignore it here.
        return [v for v in helpers.output_values(effect.aliases) if v]
    else:
        return list(effect.args)

def _get_output_reads(output_ids: list[tuple[lqp_bundle.ir.RelationId, str]]) -> list[lqp_bundle.ir.Read]:
    reads = []
    for (rel_id, name) in output_ids:
        assert isinstance(rel_id, lqp_bundle.ir.RelationId)
        output = lqp_bundle.ir.Output(name=name, relation_id=rel_id, meta=None)
        reads.append(lqp_bundle.ir.Read(read_type=output, meta=None))
    return reads

def _get_export_reads(descriptors: list[ExportDescriptor]) -> tuple[str, list, list[lqp_bundle.ir.Read]]:
    reads = []
    csv_columns = []
    col_info = []
    for descriptor in sorted(descriptors, key=lambda x: x.column_number):
        csv_columns.append(lqp_bundle.ir.ExportCSVColumn(
            column_name=descriptor.column_name,
            column_data=descriptor.relation_id,
            meta=None,
        ))
        col_info.append((descriptor.column_name, descriptor.column_type))

    # Generate a random name for the internal export path
    export_filename = "export_" + str(uuid.uuid4()).replace("-", "_")

    # Note that the engine will append the transaction id to the export path for access
    # control reasons. So the actual final path will be
    # `.../export/{export_filename}/data_{txn_id}`
    export_path = f"snowflake://APP_STATE.RAI_INTERNAL_STAGE/export/{export_filename}/data"
    export_csv_config = lqp_bundle.ir.ExportCSVConfig(
        path=export_path,
        data_columns=csv_columns,
        compression="gzip",
        partition_size=200,
        syntax_escapechar='"',  # To follow Snowflake's expected format
        meta=None,
    )
    reads.append(lqp_bundle.ir.Read(read_type=lqp_bundle.ir.Export(config=export_csv_config, meta=None), meta=None))
    return (export_filename, col_info, reads)

def _translate_to_decls(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp_bundle.ir.Declaration]:
    if is_lifted_data(rule):
        return _translate_lifted_data(ctx, rule)
    elif contains_only_declarable_constraints(rule):
        return _translate_to_constraint_decls(ctx, rule)
    elif is_algorithm_logical(rule):
        return _translate_algorithms(ctx, rule)
    else:
        return _translate_to_standard_decl(ctx, rule)

# Translate a lifted data Logical (annotated @lifted_data) into a CSVData declaration and per-column Defs.
# From
#     Logical @lifted_data
#         Data{df}(row_id, col_1, col_2, ...)
#         Logical @gnf_data("__ROW_ID__")
#             derive _data_k___row_id__(row_id)
#         Logical @gnf_data("col_1")
#             derive _data_k_col_1(row_id, col_1)
#         Logical @gnf_data("col_2")
#             derive _data_k_col_2(row_id, col_2)
#         Logical @gnf_data("col_3")
#             derive _data_k_col_3(row_id, col_3)
# To
#     (csv-data
#         (csv-locator :inline "<col_1>,<col_2>,<col_3>\n<row1>...\n<row2>...\n")
#         (csv-config :header-row 1 :delimiter "," :newline "\n" :quotechar "\"" :escapechar "\"" ...)
#         (csv-column "col_1" _data_k_col_1_col_col_1 (INT <col_1_type>))
#         (csv-column "col_2" _data_k_col_2_col_col_2 (INT <col_2_type>))
#         (csv-column "col_3" _data_k_col_3_col_col_3 (INT <col_3_type>)))
#     (def _data_k___row_id__ (row_id::INT128)
#         (exists (_row_id::INT, _v::<col_1_type>) (and (cast _row_id row_id) (atom _data_k_col_1_col_col_1 _row_id _v))))
#     (def _data_k_col_1 (row_id::INT128, col_1::<col_1_type>)
#         (exists (_row_id::INT) (and (cast _row_id row_id) (atom _data_k_col_1_col_col_1 _row_id col_1))))
#     (def _data_k_col_2 (row_id::INT128, col_2::<col_2_type>)
#         (exists (_row_id::INT) (and (cast _row_id row_id) (atom _data_k_col_2_col_col_2 _row_id col_2))))
#     (def _data_k_col_3 (row_id::INT128, col_3::<col_3_type>)
#         (exists (_row_id::INT) (and (cast _row_id row_id) (atom _data_k_col_3_col_col_3 _row_id col_3))))
#
def _translate_lifted_data(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp_bundle.ir.Declaration]:
    # Extract the Data node and @gnf_data children.
    data_node = None
    gnf_children: list[ir.Logical] = []
    for task in rule.body:
        if isinstance(task, ir.Data):
            assert data_node is None, "Expected exactly one Data node in lifted data Logical"
            data_node = task
        elif isinstance(task, ir.Logical) and has_gnf_data_annotation(task):
            gnf_children.append(task)
        else:
            raise ValueError(f"Unexpected task in lifted data Logical: {type(task)}")
    assert data_node is not None, "Missing Data node in lifted data Logical"

    # Separate @gnf_data children into the __ROW_ID__ entry and column entries.
    row_id_gnf = None
    col_gnfs: list[tuple[str, ir.Logical]] = []
    for gnf in gnf_children:
        col_name = get_gnf_column_name(gnf)
        if col_name == ROW_ID_COLUMN:
            row_id_gnf = gnf
        else:
            col_gnfs.append((col_name, gnf))
    assert row_id_gnf is not None, "Missing __ROW_ID__ @gnf_data in lifted data Logical"

    # Build inline CSV from the DataFrame.
    # Convert datetime/date columns to ISO 8601 strings before serializing,
    # because pandas to_csv uses space-separated datetime format by default
    # (e.g. "2025-01-01 10:24:10") and the engine expects ISO 8601
    # ("2025-01-01T10:24:10" for DateTime, "2025-01-01" for Date).
    # We use .isoformat() to preserve timezone information.
    df = _prepare_data_for_csv(data_node.data)
    csv_buffer = io.StringIO()
    df.to_csv(csv_buffer, index=False, lineterminator="\n")
    inline_csv = csv_buffer.getvalue()

    # Row ID and CSV index types.
    row_id_lqp_type = meta_type_to_lqp(data_node.vars[0].type)
    csv_index_type = mk_type(lqp_bundle.ir.TypeName.INT)

    # Build column CSV entries and Defs.
    csv_columns, col_defs, first_csv_col_target_id, first_csv_col_value_type = \
        _build_column_defs(ctx, col_gnfs, row_id_lqp_type, csv_index_type)

    # Build the __ROW_ID__ Def.
    row_id_def = _build_row_id_def(
        ctx, row_id_gnf, row_id_lqp_type, csv_index_type,
        first_csv_col_target_id, first_csv_col_value_type,
    )

    # Build the CSVData declaration.
    csv_data = _build_inline_csv_data(inline_csv, csv_columns)

    result: list[lqp_bundle.ir.Declaration] = []
    result.append(csv_data)
    result.append(row_id_def)
    result.extend(col_defs)
    return result


def _prepare_data_for_csv(df: "pd.DataFrame") -> "pd.DataFrame":
    """Convert datetime/date columns to ISO 8601 strings for the engine's CSV parser.

    - Timezone-aware datetimes are converted to UTC (the engine stores DateTime as UTC)
      to avoid non-standard offsets (e.g. +05:03:12) that the CSV parser cannot handle.
    - Uses .isoformat() to produce "T"-separated format (e.g. "2025-01-01T10:24:10+00:00").
    - Date columns (stored as Python date objects) are also converted via .isoformat().
    """
    from datetime import date as _date, datetime as _datetime, timezone as _tz
    result = df.copy()
    for col in result.columns:
        if pd.api.types.is_datetime64_any_dtype(result[col]):
            # Convert tz-aware datetimes to UTC to avoid non-standard tz offsets
            if hasattr(result[col].dt, 'tz') and result[col].dt.tz is not None:
                result[col] = result[col].dt.tz_convert('UTC')
            result[col] = result[col].apply(lambda x: x.isoformat() if pd.notna(x) else "")
        elif result[col].dtype == object:
            sample = result[col].dropna()
            if not sample.empty and isinstance(sample.iloc[0], _datetime):
                result[col] = result[col].apply(
                    lambda x: x.astimezone(_tz.utc).isoformat() if x is not None else ""
                )
            elif not sample.empty and isinstance(sample.iloc[0], _date):
                result[col] = result[col].apply(lambda x: x.isoformat() if x is not None else "")
    return result


def _build_inline_csv_data(
    inline_csv: str,
    gnf_columns: list[lqp_bundle.ir.GNFColumn],
) -> lqp_bundle.ir.CSVData:
    csv_locator = lqp_bundle.ir.CSVLocator(paths=[], inline_data=inline_csv.encode("utf-8"), meta=None)
    csv_config = lqp_bundle.ir.CSVConfig(
        header_row=1,
        skip=0,
        new_line="\n",
        delimiter=",",
        quotechar='"',
        escapechar='"',
        comment="",
        missing_strings=[""],
        decimal_separator=".",
        encoding="utf-8",
        compression="none",
        meta=None,
    )
    return lqp_bundle.ir.CSVData(
        locator=csv_locator,
        config=csv_config,
        columns=gnf_columns,
        asof="",
        meta=None,
    )


def _build_row_id_def(
    ctx: TranslationCtx,
    row_id_gnf: ir.Logical,
    row_id_lqp_type: lqp_bundle.ir.Type,
    csv_index_type: lqp_bundle.ir.Type,
    first_csv_col_target_id: lqp_bundle.ir.RelationId,
    first_csv_col_value_type: lqp_bundle.ir.Type,
) -> lqp_bundle.ir.Def:

    row_id_updates = collect_by_type(ir.Update, row_id_gnf)
    assert len(row_id_updates) == 1
    row_id_update = row_id_updates[0]
    row_id_bindings = _effect_bindings(row_id_update)
    row_id_projection, _ = _translate_bindings(ctx, row_id_bindings)
    row_id_rel_id = get_relation_id(ctx, row_id_update.relation, row_id_projection)

    rid_var = gen_unique_var(ctx, "row_id")
    rid_raw_var = gen_unique_var(ctx, "_row_id")
    rid_dummy_var = gen_unique_var(ctx, "_v")
    rid_shifted_var = gen_unique_var(ctx, "_row_id_shifted")

    two_literal = to_lqp_value(2, types.Int64)
    rid_subtract = mk_primitive("rel_primitive_subtract_monotype", [rid_raw_var, two_literal, rid_shifted_var])
    rid_cast = lqp_bundle.ir.Cast(input=rid_shifted_var, result=rid_var, meta=None)
    rid_atom = lqp_bundle.ir.Atom(name=first_csv_col_target_id, terms=[rid_raw_var, rid_dummy_var], meta=None)

    rid_body = mk_exists(
        [(rid_raw_var, csv_index_type), (rid_dummy_var, first_csv_col_value_type),
         (rid_shifted_var, csv_index_type)],
        mk_and([rid_subtract, rid_cast, rid_atom]),
    )

    return lqp_bundle.ir.Def(
        name=row_id_rel_id,
        body=mk_abstraction([(rid_var, row_id_lqp_type)], rid_body),
        attrs=[],
        meta=None,
    )


def _build_column_defs(
    ctx: TranslationCtx,
    col_gnfs: list[tuple[str, ir.Logical]],
    row_id_lqp_type: lqp_bundle.ir.Type,
    csv_index_type: lqp_bundle.ir.Type,
) -> tuple[list[lqp_bundle.ir.GNFColumn], list[lqp_bundle.ir.Def], lqp_bundle.ir.RelationId, lqp_bundle.ir.Type]:

    csv_columns: list[lqp_bundle.ir.GNFColumn] = []
    col_defs: list[lqp_bundle.ir.Def] = []
    first_csv_col_target_id = None
    first_csv_col_value_type = None

    for col_name, gnf_logical in col_gnfs:
        updates = collect_by_type(ir.Update, gnf_logical)
        assert len(updates) == 1, f"Expected exactly one derive in @gnf_data for column {col_name}"
        update = updates[0]
        derive_rel = update.relation

        bindings = _effect_bindings(update)
        projection, _ = _translate_bindings(ctx, bindings)
        derive_rel_id = get_relation_id(ctx, derive_rel, projection)

        assert len(derive_rel.fields) == 2, f"Expected 2-arity derive relation for column {col_name}"
        col_value_type = meta_type_to_lqp(derive_rel.fields[1].type)

        csv_col_target_id = gen_rel_id(ctx, f"{derive_rel.name}_col_{col_name}")

        if first_csv_col_target_id is None:
            first_csv_col_target_id = csv_col_target_id
            first_csv_col_value_type = col_value_type

        csv_columns.append(lqp_bundle.ir.GNFColumn(
            column_name=col_name,
            target_id=csv_col_target_id,
            types=[csv_index_type, col_value_type],
            meta=None,
        ))

        row_id_var = gen_unique_var(ctx, "row_id")
        col_var = gen_unique_var(ctx, col_name)
        raw_row_id_var = gen_unique_var(ctx, "_row_id")
        shifted_row_id_var = gen_unique_var(ctx, "_row_id_shifted")

        two_literal = to_lqp_value(2, types.Int64)
        subtract_formula = mk_primitive("rel_primitive_subtract_monotype", [raw_row_id_var, two_literal, shifted_row_id_var])
        cast_formula = lqp_bundle.ir.Cast(input=shifted_row_id_var, result=row_id_var, meta=None)
        atom_formula = lqp_bundle.ir.Atom(name=csv_col_target_id, terms=[raw_row_id_var, col_var], meta=None)

        body = mk_exists(
            [(raw_row_id_var, csv_index_type), (shifted_row_id_var, csv_index_type)],
            mk_and([subtract_formula, cast_formula, atom_formula]),
        )

        col_defs.append(lqp_bundle.ir.Def(
            name=derive_rel_id,
            body=mk_abstraction([(row_id_var, row_id_lqp_type), (col_var, col_value_type)], body),
            attrs=[],
            meta=None,
        ))

    assert first_csv_col_target_id is not None, "Expected at least one data column"
    assert first_csv_col_value_type is not None
    return csv_columns, col_defs, first_csv_col_target_id, first_csv_col_value_type


def _translate_to_constraint_decls(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp_bundle.ir.Declaration]:
    constraint_decls: list[lqp_bundle.ir.Declaration] = []
    for task in rule.body:
        if isinstance(task, ir.Logical):
            constraint_decls.extend(_translate_to_constraint_decls(ctx, task))
            continue
        else:
            assert isinstance(task, ir.Require)
            constraint_decls.extend(_translate_to_constraint_decls_from_require(ctx, task))
    return constraint_decls

def _translate_to_constraint_decls_from_require(ctx: TranslationCtx, rule: ir.Require) -> list[lqp_bundle.ir.Declaration]:
    fds = normalized_fds(rule)
    assert len(fds) > 0, f"Expected at least one functional dependency from Require node, got {len(fds)}"

    decls: list[lqp_bundle.ir.Declaration] = []
    for fd in fds:
        # check for unresolved types
        if any(types.is_any(var.type) for var in fd.keys + fd.values):
            warn(f"Ignoring FD with unresolved type: {fd}")
            continue

        lqp_typed_keys = [_translate_term(ctx, key) for key in fd.keys]
        lqp_typed_values = [_translate_term(ctx, value) for value in fd.values]
        lqp_typed_vars:list[Tuple[lqp_bundle.ir.Var, lqp_bundle.ir.Type]] = lqp_typed_keys + lqp_typed_values # type: ignore
        lqp_guard_atoms = [_translate_to_atom(ctx, atom) for atom in fd.guard]
        lqp_guard = mk_abstraction(lqp_typed_vars, mk_and(lqp_guard_atoms))
        lqp_keys:list[lqp_bundle.ir.Var] = [var for (var, _) in lqp_typed_keys] # type: ignore
        lqp_values:list[lqp_bundle.ir.Var] = [var for (var, _) in lqp_typed_values] # type: ignore
        lqp_id = utils.lqp_hash(fd.canonical_str)
        lqp_name:lqp_bundle.ir.RelationId = lqp_bundle.ir.RelationId(id=lqp_id, meta=None)

        decls.append(lqp_bundle.ir.FunctionalDependency(
            name=lqp_name,
            guard=lqp_guard,
            keys=lqp_keys,
            values=lqp_values,
            meta=None
        ))
    return decls

def _translate_algorithms(ctx: TranslationCtx, task: ir.Logical) -> list[lqp_bundle.ir.Declaration]:
    assert is_algorithm_logical(task)
    decls: list[lqp_bundle.ir.Declaration] = []
    for subtask in task.body:
        assert is_algorithm_script(subtask), "Expected all subtasks to be algorithm scripts"
        decls.extend(_translate_algorithm_script(ctx, subtask))
    return decls

def _translate_algorithm_script(ctx: TranslationCtx, alg_task: ir.Sequence) -> list[lqp_bundle.ir.Declaration]:
    assert is_algorithm_script(alg_task), "Expected Sequence @algorithm @script "

    alg_globals = _find_algorithm_global_relation_ids(ctx, alg_task)
    alg_body = _translate_script(ctx, alg_task)

    return [lqp_bundle.ir.Algorithm(global_=alg_globals, body=alg_body, meta=None)]

def _find_algorithm_global_relation_ids(ctx: TranslationCtx, alg_task: ir.Sequence) -> list[lqp_bundle.ir.RelationId]:
    result = []
    updates = collect_by_type(ir.Update, alg_task)
    for update in updates:
        if has_global_annotation(update):
            bindings = _effect_bindings(update)
            projection, _ = _translate_bindings(ctx, bindings)
            rel_id = get_relation_id(ctx, update.relation, projection)
            result.append(rel_id)
    return list(dict.fromkeys(result))

def _translate_script(ctx: TranslationCtx, task: ir.Sequence) -> lqp_bundle.ir.Script:
    assert is_script(task), "Expected a @script Sequence"

    constructs: list[lqp_bundle.ir.Construct] = []

    for subtask in task.tasks:
        if is_empty_instruction(subtask):
            constructs.append(_translate_empty_instruction(ctx, subtask))
        elif isinstance(subtask, ir.Logical):
            constructs.extend(_translate_instruction(ctx, subtask))
        elif isinstance(subtask, ir.Break):
            constructs.append(_translate_break_instruction(ctx, subtask))
        elif is_while_loop(subtask):
            constructs.append(_translate_while_loop(ctx, subtask))
        else:
            raise Exception(f"Unsupported script instruction: {subtask}")

    return lqp_bundle.ir.Script(constructs=constructs, meta=None)

def _translate_while_loop(ctx: TranslationCtx, task: ir.Loop) -> lqp_bundle.ir.Loop:
    assert is_while_loop(task), "Expected a @while Loop"
    assert len(task.iter) == 0, "Temporalized loops not supported"

    while_script_task = task.body
    assert is_while_script(while_script_task), "The body of a @while Loop must be a @while @script Sequence"
    body_script = _translate_script(ctx, while_script_task)

    # No init instructions in the translation of PyRel Loops to to LQP loops
    return lqp_bundle.ir.Loop(init=[], body=body_script, meta=None)

def _translate_break_instruction(ctx: TranslationCtx, task: ir.Break) -> lqp_bundle.ir.Construct:
    body = _translate_to_formula(ctx, task.check)

    ctx.break_rule_counter += 1

    rel_id = gen_rel_id(ctx, "break_cond_" + str(ctx.break_rule_counter))
    return lqp_bundle.ir.Break(
        name = rel_id,
        body = mk_abstraction([], body),
        attrs = [],
        meta = None,
    )

def _translate_empty_instruction(ctx: TranslationCtx, rule: ir.Logical) -> lqp_bundle.ir.Instruction:
    assert is_empty_instruction(rule)
    updates = collect_by_type(ir.Update, rule)
    assert len(updates) == 1
    update = updates[0]
    bindings = _effect_bindings(update)

    # We need  to make sure that variable names have a leading underscore
    normalized_bindings:Sequence[ir.Var] = []
    for v in bindings:
        assert isinstance(v, ir.Var)
        if not v.name.startswith("_"):
            v = ir.Var(v.type, "_" + v.name)
        normalized_bindings.append(v)

    projection, eqs = _translate_bindings(ctx, normalized_bindings)
    assert len(eqs) == 0
    rel_id = get_relation_id(ctx, update.relation, projection)
    abstraction = mk_abstraction(projection, mk_or([])) # empty body = false
    return lqp_bundle.ir.Assign(name = rel_id, body = abstraction, attrs = [], meta = None)

def _translate_instruction(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp_bundle.ir.Instruction]:
    effects = collect_by_type((ir.Update, ir.Output), rule)

    if len(effects) == 0:
        return []

    conjuncts = _translate_to_formula(ctx, rule)
    res = []
    for effect in effects:
        assert isinstance(effect, ir.Update), f"Got an effect of type {type(effect)} in a loop, which is invalid."

        bindings = _effect_bindings(effect)
        projection, eqs = _translate_bindings(ctx, bindings)

        eqs.append(conjuncts)
        new_body = mk_and(eqs)

        rel_id = get_relation_id(ctx, effect.relation, projection)
        abstraction = mk_abstraction(projection, new_body)

        annotations = effect.annotations | effect.relation.annotations
        attrs = _translate_annotations(annotations)

        upsert = get_upsert_annotation(effect)
        monoid = get_monoid_annotation(effect)
        monus = get_monus_annotation(effect)

        if has_assign_annotation(effect):
            res.append(lqp_bundle.ir.Assign(name = rel_id, body = abstraction, attrs = attrs, meta = None))
        elif upsert is not None:
            res.append(lqp_bundle.ir.Upsert(value_arity=get_arity(upsert), name = rel_id, body = abstraction, attrs = attrs, meta = None))
        elif monoid is not None:
            res.append(lqp_bundle.ir.MonoidDef(
                value_arity=get_arity(monoid),
                monoid=construct_monoid(monoid),
                name = rel_id,
                body = abstraction,
                attrs = attrs,
                meta = None
            ))
        elif monus is not None:
            res.append(lqp_bundle.ir.MonusDef(
                value_arity=get_arity(monus),
                monoid=construct_monoid(monus),
                name = rel_id,
                body = abstraction,
                attrs = attrs,
                meta = None
            ))

    return res

def _translate_to_standard_decl(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp_bundle.ir.Declaration]:
    effects = collect_by_type((ir.Output, ir.Update), rule)

    # TODO: should this ever actually come in as input?
    if len(effects) == 0:
        return []

    conjuncts = _translate_to_formula(ctx, rule)
    return [_translate_effect(ctx, effect, conjuncts) for effect in effects]

def _translate_annotations(annotations: FrozenOrderedSet[ir.Annotation]) -> list[lqp_bundle.ir.Attribute]:
    attributes = []
    for annotation in annotations:
        if annotation.relation.name in supported_lqp_annotations:
            attr = _translate_annotation(annotation)
            if attr is not None:
                attributes.append(attr)
    return attributes

def _translate_annotation(annotation: ir.Annotation) -> Optional[lqp_bundle.ir.Attribute]:
    if any(not isinstance(a, ir.Literal) for a in annotation.args):
        warn("LQP currently ignores annotation parameters with non-literal values")
        return None
    if annotation.relation.name == builtins.incremental_maintenance.name:
        return _translate_ivm_annotation(annotation)
    else:
        # Convert literal arguments to LQP values
        args = []
        for a in annotation.args:
            assert isinstance(a, ir.Literal)
            args.append(mk_value(a.value))
        return mk_attribute(annotation.relation.name, args)

def _translate_ivm_annotation(annotation: ir.Annotation) -> Optional[lqp_bundle.ir.Attribute]:
    if len(annotation.args) == 0:
        warn("incremental_maintenance annotation requires a value argument.")
        return None
    arg = annotation.args[0]
    assert isinstance(arg, ir.Literal)
    value = arg.value
    if value not in ("all", "off"):
        warn(f"Invalid value for incremental_maintenance annotation: {value}. Expected 'all' or 'off'.")
        return None
    else:
        if value == "all":
            return mk_attribute("must_try_maintain", [])
        elif value == "off":
            return mk_attribute("nomaintain", [])

# Translates an effect (export, output, or update) into the corresponding def. Note that
# this method only generates the def and not the LQP read operation, which is added later by
# `_get_output_reads` and `_get_export_reads`.
def _translate_effect(ctx: TranslationCtx, effect: Union[ir.Output, ir.Update], body: lqp_bundle.ir.Formula) -> lqp_bundle.ir.Declaration:
    bindings = _effect_bindings(effect)

    if isinstance(effect, ir.Output):
        projection, eqs, suffix = _translate_output_bindings(ctx, bindings)
        meta_id = effect.id

        if helpers.is_export(effect):
            def_name = "export_relation" + suffix
        else:
            def_name = "output" + suffix

        def_name = ctx.output_names.get_name_by_id(meta_id, def_name)
        rel_id = get_output_id(ctx, def_name, meta_id)
    else:
        projection, eqs = _translate_bindings(ctx, bindings)
        def_name = effect.relation.name
        rel_id = get_relation_id(ctx, effect.relation, projection)

    eqs.append(body)
    new_body = mk_and(eqs)

    # Context bookkeeping for exports and outputs
    if helpers.is_export(effect):
        # The row id is the first n-1 elements, and the actual data is the last element. Its
        # type is stored in the first element of the tuple.
        col_type = projection[-1][1]
        _col_num_match = re.search(r"export_relation_col([0-9]+)", def_name)
        assert _col_num_match, f"Could not find column number in suffix: {def_name}"
        col_num = int(_col_num_match.group(1))
        col_name = f"col{col_num}"
        if isinstance(effect, ir.Output) and len(effect.aliases) > 0:
            aliases_list = list(effect.aliases)
            col_name = aliases_list[-1][0]

        # Extract export FQN from annotations
        export_fqn_str = extract_export_fqn(effect.annotations)

        ctx.export_descriptors.append(ExportDescriptor(
            relation_id=rel_id,
            column_name=col_name,
            column_number=col_num,
            column_type=col_type,
            export_fqn=export_fqn_str,
        ))
    elif isinstance(effect, ir.Output):
        ctx.output_ids.append((rel_id, def_name))

    # First we collect annotations on the effect itself, e.g. from something like
    # `select(...).annotate(...)`.
    annotations = effect.annotations
    if isinstance(effect, ir.Update):
        # Then we translate annotations on the relation itself, e.g.
        # ```
        # Bar.foo = model.Relationship(...)
        # Bar.foo.annotate(...)
        # ```
        annotations = annotations | effect.relation.annotations

    return lqp_bundle.ir.Def(
        name = rel_id,
        body = mk_abstraction(projection, new_body),
        attrs = _translate_annotations(annotations),
        meta = None,
    )

def _translate_output_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp_bundle.ir.Var, lqp_bundle.ir.Type]], list[lqp_bundle.ir.Formula], str]:
    symbol_literals = []
    non_symbols = []
    for binding in bindings:
        if isinstance(binding, ir.Literal) and binding.type == types.Symbol:
            symbol_literals.append(binding.value)
        else:
            non_symbols.append(binding)
    projection, eqs = _translate_bindings(ctx, non_symbols)
    if len(symbol_literals) > 0:
        name_suffix = "_"
        name_suffix += "_".join(symbol_literals)
    else:
        name_suffix = ""

    return projection, eqs, name_suffix

def _translate_rank(ctx: TranslationCtx, rank: ir.Rank, body: lqp_bundle.ir.Formula) -> lqp_bundle.ir.Formula:
    # Ascending rank is constructed using rel_primitive_sort. If a limit is added to an
    # ascending rank we can use rel_primitive_top for an efficient evaluation.
    #
    # Descending rank is constructed as an ascending sort, then the ascending rank is
    # subtracted from a count of the elements (plus 1 so that we still start at 1).
    # Adding a limit to a descending rank is done by adding a filter of rank <= limit.

    # Limits are the sort plus a filter on rank <= limit.
    if all(o for o in rank.arg_is_ascending):
        ascending = True
    elif all(not o for o in rank.arg_is_ascending):
        ascending = False
    else:
        raise Exception("Mixed orderings in rank are not supported yet.")

    # Filter out the group-by variables, since they are introduced outside the rank.
    input_args, input_eqs = _translate_bindings(ctx, list(rank.args))
    introduced_meta_projs = [arg for arg in rank.projection if arg not in rank.group and arg not in rank.args]
    projected_args, projected_eqs = _translate_bindings(ctx, list(introduced_meta_projs))

    # rank expects an Int128 result, but the primitive will return an Int64 result.
    # we need to set up an intermediary variable to hold the Int64 result, and a cast
    # to convert it to Int128.
    result_var, _ = _translate_term(ctx, rank.result)
    # The primitive will return an Int64 result, so we need a var to hold the intermediary.
    result_64_var = gen_unique_var(ctx, "v_rank")
    result_64_type = mk_type(lqp_bundle.ir.TypeName.INT)

    cast = lqp_bundle.ir.Cast(input=result_64_var, result=result_var, meta=None)

    body = mk_and([body] + input_eqs + projected_eqs)
    abstr_args = input_args + projected_args

    if ascending:
        ranker = _translate_ascending_rank(ctx, rank.limit, result_64_var, body, abstr_args)
    else:
        ranker = _translate_descending_rank(ctx, rank.limit, result_64_var, body, abstr_args)

    return mk_exists([(result_64_var, result_64_type)], mk_and([ranker, cast]))

def _translate_descending_rank(ctx: TranslationCtx, limit: int, result: lqp_bundle.ir.Var, body: lqp_bundle.ir.Formula, abstr_args) -> lqp_bundle.ir.Formula:
    result_var = result
    result_type = mk_type(lqp_bundle.ir.TypeName.INT)

    # Rename abstracted args in the body to new variable names
    var_map = {var.name: gen_unique_var(ctx, 't_' + var.name) for (var, _) in abstr_args}
    body = utils.rename_vars_formula(body, var_map)
    new_abstr_args = [(var_map[var.name], typ) for (var, typ) in abstr_args]

    # Construct a conjunction of the ranking, a counter for the body, a subtraction
    # of the rank from the count and an addition of 1. Wrap this in an abstraction.
    count_res = gen_unique_var(ctx, "count_res")

    # Add one to the count to account for the rank starting at 1.
    one, one_eq = constant_to_var(ctx, to_lqp_value(1, types.Int64), "one")
    one_bigger = gen_unique_var(ctx, "one_bigger")
    addition = mk_primitive("rel_primitive_add_monotype", [count_res, one, one_bigger])

    # Subtract the rank from the count + 1
    asc_rank = gen_unique_var(ctx, "asc_rank")
    subtraction = mk_primitive("rel_primitive_subtract_monotype", [one_bigger, result_var, asc_rank])

    # Construct the ranking
    desc_ranking_terms = [asc_rank] + [v[0] for v in abstr_args]
    ranking = lqp_bundle.ir.FFI(
        meta=None,
        name="rel_primitive_sort",
        args=[mk_abstraction(new_abstr_args, body)],
        terms=desc_ranking_terms,
    )

    # Count the number of rows in the body
    count_type = meta_type_to_lqp(types.Int64)
    count_var, count_eq = constant_to_var(ctx, to_lqp_value(1, types.Int64), "counter")
    desc_body = mk_and([body, count_eq])
    aggr_abstr_args = new_abstr_args + [(count_var, count_type)]
    count_aggr = lqp_bundle.ir.Reduce(
        op=lqp_operator(
            ctx,
            "count",
            "count",
            mk_type(lqp_bundle.ir.TypeName.INT)
        ),
        body=mk_abstraction(aggr_abstr_args, desc_body),
        terms=[count_res],
        meta=None
    )

    # Bring it all together and do the maths.
    ranking = mk_exists(
        vars=[
            (asc_rank, result_type),
            (count_res, result_type),
            (one, result_type),
            (one_bigger, result_type)
        ],
        value=mk_and([ranking, count_aggr, one_eq, addition, subtraction])
    )

    # If there is a limit, we need to add a filter to the ranking.
    # Wrap with a rank <= limit
    if limit != 0:
        limit_term, _ = _translate_term(ctx, ir.Literal(types.Int64, limit))
        limiter = mk_primitive("rel_primitive_lt_eq_monotype", [result_var, limit_term])
        ranking = mk_and([ranking, limiter])

    return ranking

def _translate_ascending_rank(ctx: TranslationCtx, limit: int, result_var: lqp_bundle.ir.Var, body: lqp_bundle.ir.Formula, abstr_args) -> lqp_bundle.ir.Formula:
    terms = [result_var] + [v[0] for v in abstr_args]

    # Rename abstracted args in the body to new variable names
    var_map = {var.name: gen_unique_var(ctx, 't_' + var.name) for (var, _) in abstr_args}
    body = utils.rename_vars_formula(body, var_map)
    new_abstr_args = [(var_map[var.name], typ) for (var, typ) in abstr_args]
    sort_abstr = mk_abstraction(new_abstr_args, body)

    if limit == 0:
        return lqp_bundle.ir.FFI(
            meta=None,
            name="rel_primitive_sort",
            args=[sort_abstr],
            terms=terms,
        )
    else:
        limit_type = meta_type_to_lqp(types.Int64)
        limit_var, limit_eq = constant_to_var(ctx, to_lqp_value(limit, types.Int64), "limit")
        limit_abstr = mk_abstraction([(limit_var, limit_type)], limit_eq)
        return lqp_bundle.ir.FFI(
            meta=None,
            name="rel_primitive_top",
            args=[sort_abstr, limit_abstr],
            terms=terms,
        )

def _rename_shadowed_abstraction_vars(
    ctx: TranslationCtx,
    aggr: ir.Aggregate,
    abstr_args: list[Tuple[lqp_bundle.ir.Var, lqp_bundle.ir.Type]],
    body_conjs: list[lqp_bundle.ir.Formula]
) -> list[Tuple[lqp_bundle.ir.Var, lqp_bundle.ir.Type]]:
    """
    Rename abstraction variables that shadow group-by variables.

    This can happen when the same variable appears in both aggr.group and as an input
    to the aggregation, e.g., min(Person.age).per(Person.age). The group-by variables
    are in the outer scope, while the abstraction parameters are in the inner scope,
    so we need different names to avoid shadowing.
    """
    # Get the LQP names of group-by variables
    group_var_names = set()
    for group_var in aggr.group:
        lqp_var = _translate_var(ctx, group_var)
        group_var_names.add(lqp_var.name)

    # Rename any abstraction parameters that conflict with group-by variables
    renamed_abstr_args = []
    for var, typ in abstr_args:
        if var.name in group_var_names:
            # This variable shadows a group-by variable, so rename it
            fresh_var = gen_unique_var(ctx, var.name)
            # Add an equality constraint: fresh_var == var
            # var is a free variable referring to the outer scope group-by variable
            body_conjs.append(mk_primitive("rel_primitive_eq", [fresh_var, var]))
            renamed_abstr_args.append((fresh_var, typ))
        else:
            renamed_abstr_args.append((var, typ))

    return renamed_abstr_args

def _translate_aggregate(ctx: TranslationCtx, aggr: ir.Aggregate, body: lqp_bundle.ir.Formula) -> Union[lqp_bundle.ir.Reduce, lqp_bundle.ir.Formula]:
    # TODO: handle this properly
    aggr_name = aggr.aggregation.name
    supported_aggrs = ("sum", "count", "avg", "min", "max", "rel_primitive_solverlib_ho_appl")
    assert aggr_name in supported_aggrs, f"only support {supported_aggrs} for now, not {aggr.aggregation.name}"

    meta_output_terms = []
    meta_input_terms = []

    for (field, arg) in zip(aggr.aggregation.fields, aggr.args):
        if field.input:
            meta_input_terms.append(arg)
        else:
            meta_output_terms.append(arg)

    output_terms = [_translate_term(ctx, term) for term in meta_output_terms]
    output_vars = [term[0] for term in output_terms]

    body_conjs = [body]
    input_args, input_eqs = _translate_bindings(ctx, meta_input_terms)

    # TODO: Can this safely be applied to all aggregates?
    if aggr_name in ("sum", "min", "max"):
        assert len(output_terms) == 1, f"{aggr_name} expects a single output variable"
        assert len(meta_input_terms) == 1, f"{aggr_name} expects a single input variable"
        assert isinstance(meta_output_terms[0], ir.Var)
        assert input_args[0][1] == output_terms[0][1], f"{aggr_name}({input_args[0][1].type_name}) had output type of {output_terms[0][1].type_name}"

    # Filter out the group-by variables, since they are introduced outside the aggregation.
    # Input terms are added later below.
    introduced_meta_projs = [arg for arg in aggr.projection if arg not in aggr.group and arg not in meta_input_terms]
    projected_args, projected_eqs = _translate_bindings(ctx, list(introduced_meta_projs))
    body_conjs.extend(input_eqs)
    body_conjs.extend(projected_eqs)
    abstr_args: list[Tuple[lqp_bundle.ir.Var, lqp_bundle.ir.Type]] = projected_args + input_args

    # Rename abstraction variables that shadow group-by variables
    abstr_args = _rename_shadowed_abstraction_vars(ctx, aggr, abstr_args, body_conjs)

    if aggr_name == "count":
        assert len(output_terms) == 1, "Count and avg expect a single output variable"
        assert isinstance(meta_output_terms[0], ir.Var)
        # Count sums up "1" for each row. We use the expected output type for the type
        # of the count variable.
        typ = meta_type_to_lqp(meta_output_terms[0].type)
        one_var, eq = constant_to_var(ctx, to_lqp_value(1, meta_output_terms[0].type), "one")
        body_conjs.append(eq)
        abstr_args.append((one_var, typ))

    # Average needs to wrap the reduce in Exists(Conjunction(Reduce, div))
    if aggr_name == "avg":
        assert len(output_vars) == 1, "avg should only have one output variable"
        output_var = output_vars[0]

        # Count sums up "1" for each row. We make the reasonably safe assumption that there
        # are less than 2^31 rows in the body.
        count_type = meta_type_to_lqp(types.Int64)
        one_var, eq = constant_to_var(ctx, to_lqp_value(1, types.Int64), "one")
        body_conjs.append(eq)
        abstr_args.append((one_var, count_type))

        # The average will produce two output variables: sum and count.
        sum_result = gen_unique_var(ctx, "sum")
        count_result = gen_unique_var(ctx, "count")

        # Second to last is the variable we're summing over.
        (sum_var, sum_type) = abstr_args[-2]
        body = mk_and(body_conjs)

        result = lqp_bundle.ir.Reduce(
            op=lqp_avg_op(ctx, aggr.aggregation.name, sum_var.name, sum_type),
            body=mk_abstraction(abstr_args, body),
            terms=[sum_result, count_result],
            meta=None,
        )

        if sum_type == count_type:
            div = mk_primitive("rel_primitive_divide_monotype", [sum_result, count_result, output_var])
            conjunction = mk_and([result, div])

            # Finally, we need to wrap everything in an `exists` to project away the sum and
            # count variables and only keep the result of the division.
            result_terms = [(sum_result, sum_type), (count_result, count_type)]
        else:
            # If the sum type and count type don't match, we need to cast the count
            count_casted = gen_unique_var(ctx, "count_casted")
            count_cast = lqp_bundle.ir.Cast(input=count_result, result=count_casted, meta=None)

            div = mk_primitive("rel_primitive_divide_monotype", [sum_result, count_casted, output_var])
            conjunction = mk_and([result, count_cast, div])

            # Finally, we need to wrap everything in an `exists` to project away the sum and
            # count variables and only keep the result of the division.
            result_terms = [(sum_result, sum_type), (count_result, count_type), (count_casted, sum_type)]

        return mk_exists(result_terms, conjunction)

    # `input_args` hold the types of the input arguments, but they may have been modified
    # if we're dealing with a count, so we use `abstr_args` to find the type.
    (aggr_arg, aggr_arg_type) = abstr_args[-1]
    body = mk_and(body_conjs)

    # Group-bys do not need to be handled at all, since they are introduced outside already
    reduce = lqp_bundle.ir.Reduce(
        op=lqp_operator(ctx, aggr.aggregation.name, aggr_arg.name, aggr_arg_type),
        body=mk_abstraction(abstr_args, body),
        terms=output_vars,
        meta=None
    )
    return reduce

def _translate_to_formula(ctx: TranslationCtx, task: ir.Task) -> lqp_bundle.ir.Formula:
    if isinstance(task, ir.Logical):
        # For aggregates and ranks, the expected format is:
        #
        # Logical
        #   body_task1
        #   body_task2
        #   ...
        #   aggregate/rank task
        #
        # If we see that the Logical is in this format, it should be translated as an
        # aggregate/rank node.
        groups = group_tasks(task.body, {
            "aggregates": ir.Aggregate,
            "ranks": ir.Rank,
        })

        aggregates = groups.get("aggregates", OrderedSet[ir.Task]())
        ranks = groups.get("ranks", OrderedSet[ir.Task]())

        if aggregates or ranks:
            conjuncts = []
            body = mk_and([_translate_to_formula(ctx, t) for t in task.body])
            for aggr in aggregates:
                assert(isinstance(aggr, ir.Aggregate))
                conjuncts.append(_translate_aggregate(ctx, aggr, body))
            for rank in ranks:
                assert(isinstance(rank, ir.Rank))
                conjuncts.append(_translate_rank(ctx, rank, body))
        else:
            # If there are no aggregates or ranks, translate as a normal conjunction
            conjuncts = [_translate_to_formula(ctx, child) for child in task.body]
        return mk_and(conjuncts)
    elif isinstance(task, ir.Lookup):
        return _translate_to_atom(ctx, task)
    elif isinstance(task, ir.Not):
        return lqp_bundle.ir.Not(arg=_translate_to_formula(ctx, task.task), meta=None)
    elif isinstance(task, ir.Exists):
        lqp_vars, conjuncts = _translate_bindings(ctx, list(task.vars))
        conjuncts.append(_translate_to_formula(ctx, task.task))
        return mk_exists(lqp_vars, mk_and(conjuncts))
    elif isinstance(task, ir.Construct):
        assert len(task.values) >= 1, "Construct should have at least one value"
        terms = [_translate_term(ctx, arg) for arg in task.values]
        result_term = _translate_term(ctx, task.id_var)
        terms.append(result_term)
        assert result_term[1].type_name == lqp_bundle.ir.TypeName.UINT128, \
            f"Attempting to store a {task.id_var} in a type `{result_term[1].type_name}`"

        return mk_primitive("rel_primitive_hash_tuple_uint128", [v for v, _ in terms])
    elif isinstance(task, ir.Union):
        disjs = [_translate_to_formula(ctx, child) for child in task.tasks]
        return mk_or(disjs)
    elif isinstance(task, (ir.Output, ir.Update)):
        # Nothing to do here, handled in _translate_to_decls
        return mk_and([])
    elif isinstance(task, (ir.Aggregate, ir.Rank)):
        # Nothing to do here, handled at the Logical level
        return mk_and([])
    elif isinstance(task, ir.Data):
        # Data nodes should be lifted before translation to LQP
        raise NotImplementedError("Data nodes should be lifted before translation to LQP")
    else:
        raise NotImplementedError(f"Unknown task type (formula): {type(task)}")

# Only used for translating terms on relatoms, which can be specialized values.
def _translate_relterm(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp_bundle.ir.RelTerm, lqp_bundle.ir.Type]:
    if isinstance(term, ir.Literal) and term.type == types.Symbol:
        if isinstance(term.value, str):
            value = mk_value(term.value)
            return mk_specialized_value(value), meta_type_to_lqp(types.String)
        elif isinstance(term.value, int):
            value = mk_value(term.value)
            return mk_specialized_value(value), meta_type_to_lqp(types.Int64)
        else:
            raise NotImplementedError(f"Cannot specialize literal of type {type(term.value)}")
    return _translate_term(ctx, term)

def _translate_term(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp_bundle.ir.Term, lqp_bundle.ir.Type]:
    if isinstance(term, ir.ScalarType):
        # TODO: ScalarType is not like other terms, should be handled separately.
        return to_lqp_value(term.name, types.String), meta_type_to_lqp(types.String)
    elif isinstance(term, ir.Var):
        t = meta_type_to_lqp(term.type)
        return _translate_var(ctx, term), t
    else:
        assert isinstance(term, ir.Literal), f"Cannot translate value {term!r} of type {type(term)} to LQP Term; neither Var nor Literal."
        v = to_lqp_value(term.value, term.type)
        return v, meta_type_to_lqp(term.type)

def _translate_to_atom(ctx: TranslationCtx, task: ir.Lookup) -> lqp_bundle.ir.Formula:
    if task.relation == builtins.cast:
        assert len(task.args) == 3, f"expected three terms for {task.relation.name}, got {len(task.args)}"

        terms = []
        for arg in task.args[1:]:
            term, _ = _translate_relterm(ctx, arg)
            terms.append(term)
        return lqp_bundle.ir.Cast(input=terms[0], result=terms[1], meta=None)

    if task.relation == builtins.join:
        return _translate_join(ctx, task)
    elif task.relation == builtins.infomap:
        return _translate_infomap(ctx, task)
    elif task.relation == builtins.louvain:
        return _translate_louvain(ctx, task)
    elif task.relation == builtins.label_propagation:
        return _translate_label_propagation(ctx, task)

    terms = []
    term_types = []
    for arg in task.args:
        # Handle varargs, which come wrapped in a tuple.
        if isinstance(arg, tuple):
            for vararg in arg:
                term, ty = _translate_relterm(ctx, vararg)
                terms.append(term)
                term_types.append(ty)
        else:
            term, ty = _translate_relterm(ctx, arg)
            terms.append(term)
            term_types.append(ty)

    if builtins.is_pragma(task.relation):
        lqp_name = pragma_to_lqp_name(task.relation.name)
        return mk_pragma(lqp_name, terms)

    if builtins.is_builtin(task.relation):
        return build_primitive(task.relation.name, terms, term_types)

    if helpers.is_external(task.relation):
        return lqp_bundle.ir.RelAtom(name=task.relation.name, terms=terms, meta=None)

    projection, _ = _translate_bindings(ctx, list(task.args))
    rid = get_relation_id(ctx, task.relation, projection)
    return lqp_bundle.ir.Atom(name=rid, terms=terms, meta=None)


def get_relation_id(ctx: TranslationCtx, relation: ir.Relation, projection: list[Tuple[lqp_bundle.ir.Var, lqp_bundle.ir.Type]] = []) -> lqp_bundle.ir.RelationId:
    types = "_".join([str(t.type_name) for (_, t) in projection])
    if types:
        types = "_" + types
    if relation.id in ctx.def_names.id_to_name:
        unique_name = ctx.def_names.id_to_name[relation.id]
    else:
        name = helpers.python_qualified_name(relation)
        unique_name = ctx.def_names.get_name_by_id(relation.id, name)

    return utils.gen_rel_id(ctx, unique_name, types)

def get_output_id(ctx: TranslationCtx, orig_name: str, metamodel_id: int) -> lqp_bundle.ir.RelationId:
    unique_name = ctx.output_names.get_name_by_id(metamodel_id, orig_name)
    return utils.gen_rel_id(ctx, unique_name)

def _translate_bindings(ctx: TranslationCtx, bindings: Sequence[ir.Value]) -> Tuple[list[Tuple[lqp_bundle.ir.Var, lqp_bundle.ir.Type]], list[lqp_bundle.ir.Formula]]:
    lqp_vars = []
    conjuncts = []
    for binding in bindings:
        lqp_var, typ, eq = binding_to_lqp_var(ctx, binding)
        lqp_vars.append((lqp_var, typ))
        if eq is not None:
            conjuncts.append(eq)

    return lqp_vars, conjuncts

def binding_to_lqp_var(ctx: TranslationCtx, binding: ir.Value) -> Tuple[lqp_bundle.ir.Var, lqp_bundle.ir.Type, Union[None, lqp_bundle.ir.Formula]]:
    if isinstance(binding, ir.Var):
        var, typ = _translate_term(ctx, binding)
        assert isinstance(var, lqp_bundle.ir.Var)
        return var, typ, None
    elif isinstance(binding, ir.Literal):
        lqp_value = to_lqp_value(binding.value, binding.type)
        var, formula = constant_to_var(ctx, lqp_value)
        return var, meta_type_to_lqp(binding.type), formula
    else:
        raise Exception(f"Unsupported binding type: {type(binding)}")

def to_lqp_value(value: ir.PyValue, value_type: ir.Type) -> lqp_bundle.ir.Value:
    typ = meta_type_to_lqp(value_type)

    # Ensure int values match the requested integer type.
    if typ.type_name == lqp_bundle.ir.TypeName.INT and isinstance(value, int):
        assert lqp_types.INT_MIN <= value <= lqp_types.INT_MAX, f"{value} out of range for a 64-bit INT value"
        val = value
    elif typ.type_name == lqp_bundle.ir.TypeName.INT128 and isinstance(value, int):
        assert lqp_types.INT128_MIN <= value <= lqp_types.INT128_MAX, f"{value} out of range for an INT128 value"
        val = lqp_bundle.ir.Int128Value(value=value, meta=None)
    elif typ.type_name == lqp_bundle.ir.TypeName.UINT128 and isinstance(value, int):
        assert lqp_types.UINT128_MIN <= value <= lqp_types.UINT128_MAX, f"{value} out of range for an UINT128 value"
        val = lqp_bundle.ir.UInt128Value(value=value, meta=None)
    elif typ.type_name == lqp_bundle.ir.TypeName.FLOAT and isinstance(value, float):
        val = value
    elif typ.type_name == lqp_bundle.ir.TypeName.STRING and isinstance(value, str):
        val = value
    elif typ.type_name == lqp_bundle.ir.TypeName.DECIMAL and isinstance(value, (int, float, PyDecimal)):
        precision = typ.parameters[0].value
        scale = typ.parameters[1].value
        assert isinstance(precision, int) and isinstance(scale, int)
        val = lqp_bundle.ir.DecimalValue(precision=precision, scale=scale, value=PyDecimal(value), meta=None)
    elif typ.type_name == lqp_bundle.ir.TypeName.DATE and isinstance(value, date):
        val = lqp_bundle.ir.DateValue(value=value, meta=None)
    elif typ.type_name == lqp_bundle.ir.TypeName.DATETIME and isinstance(value, datetime):
        utc_value = value.astimezone(timezone.utc) if value.tzinfo is not None else value # Convert to UTC cf. Iceberg
        val = lqp_bundle.ir.DateTimeValue(value=utc_value, meta=None)
    elif typ.type_name == lqp_bundle.ir.TypeName.BOOLEAN and isinstance(value, bool):
        val = lqp_bundle.ir.BooleanValue(value=value, meta=None)
    else:
        raise Exception(f"Unsupported type for constant: {typ.type_name} with value {value} of type {type(value)}")

    return mk_value(val)

def constant_to_var(ctx: TranslationCtx, value: lqp_bundle.ir.Value, name_hint: str = "cvar") -> Tuple[lqp_bundle.ir.Var, lqp_bundle.ir.Formula]:
    var = gen_unique_var(ctx, name_hint)
    eq = mk_primitive("rel_primitive_eq", [var, value])
    return var, eq

def _extract_pyrel_error_ids(ctx: TranslationCtx, model: ir.Model) -> list[Tuple[lqp_bundle.ir.RelationId, str]]:
    effects = collect_by_type(ir.Update, model)
    pyrel_error_attrs = []
    # We use a separate counter here to avoid modifying the current model.
    i = 1
    for effect in effects:
        if "pyrel_error_attrs" not in effect.relation.name:
            continue
        projection, _ = _translate_bindings(ctx, _effect_bindings(effect))
        rel_id = get_relation_id(ctx, effect.relation, projection)
        name = f"{effect.relation.name}_{i}"
        pyrel_error_attrs.append((rel_id, name))
        i += 1

    return pyrel_error_attrs

# Translate a relation reference into an abstraction over its fields.
def _translate_relation_ref(ctx: TranslationCtx, relation: ir.Relation) -> lqp_bundle.ir.Abstraction:
    projection = []
    for field in relation.fields:
        var = gen_unique_var(ctx, field.name)
        typ = meta_type_to_lqp(field.type)
        projection.append((var, typ))

    rid = get_relation_id(ctx, relation, projection)
    atom = lqp_bundle.ir.Atom(name=rid, terms=[var for (var, _) in projection], meta=None)
    return mk_abstraction(projection, atom)

# Common translation logic for graph algorithms.
# task.args[0] : normalized weight list (int64, int64, float)
# task.args[1] : normalized node count (relation or constant int64)
# task.args[2] : normalized edge count (relation or constant int64)
# task.args[3:-3] : algorithm parameters (var or constant)
# task.args[-3] : diagnostic info
# task.args[-2] : node index
# task.args[-1] : community ident
def _translate_graph_common(name: str, ctx: TranslationCtx, task: ir.Lookup):
    abstractions = []

    assert isinstance(task.args[0], ir.Relation), \
        f"Expected relation as first arg to {name}, got {task.args[0]}:{type(task.args[0])}"
    abstractions.append(_translate_relation_ref(ctx, task.args[0]))

    # Allow constant args for node and edge count
    for arg in task.args[1:3]:
        if isinstance(arg, ir.Relation):
            abst = _translate_relation_ref(ctx, arg)
            typ = abst.vars[0][1]
            assert typ.type_name == lqp_bundle.ir.TypeName.INT, \
                f"Expected Int64 types for node and edge counts, got type {typ.type_name}"
            abstractions.append(abst)
        else:
            var, typ, eq = binding_to_lqp_var(ctx, arg)
            assert eq is not None, \
                f"Expected equality formula for {name} arg {arg}:{type(arg)}"
            abstractions.append(mk_abstraction([(var, typ)], mk_and([eq])))

    for arg in task.args[3:-3]:
        var, typ, eq = binding_to_lqp_var(ctx, arg)
        if eq:
            abstractions.append(mk_abstraction([(var, typ)], mk_and([eq])))
        else:
            print(f"Primitive graph algorithm arg without eq:\n var:{var}, typ:{typ}\n arg:{arg}:{type(arg)}")
            abstractions.append(mk_abstraction([(var, typ)], mk_and([])))

    terms = []
    for arg in task.args[-3:]:
        term, _ = _translate_relterm(ctx, arg)
        terms.append(term)

    return (abstractions, terms)

def _translate_infomap(ctx: TranslationCtx, task: ir.Lookup) -> lqp_bundle.ir.Formula:
    abstractions, terms = _translate_graph_common("infomap", ctx, task)

    return lqp_bundle.ir.FFI(
        meta=None,
        name="rel_primitive_infomap",
        args=abstractions,
        terms=terms,
    )

def _translate_louvain(ctx: TranslationCtx, task: ir.Lookup) -> lqp_bundle.ir.Formula:
    abstractions, terms = _translate_graph_common("louvain", ctx, task)

    return lqp_bundle.ir.FFI(
        meta=None,
        name="rel_primitive_louvain",
        args=abstractions,
        terms=terms,
    )

def _translate_label_propagation(ctx: TranslationCtx, task: ir.Lookup) -> lqp_bundle.ir.Formula:
    abstractions, terms = _translate_graph_common("label_propagation", ctx, task)

    return lqp_bundle.ir.FFI(
        meta=None,
        name="rel_primitive_async_label_propagation",
        args=abstractions,
        terms=terms,
    )

# Hard-coded implementation of Rel's string_join
def _translate_join(ctx: TranslationCtx, task: ir.Lookup) -> lqp_bundle.ir.Formula:
    assert len(task.args) == 3
    (strs, separator, target) = task.args
    assert isinstance(separator, ir.Node)
    assert isinstance(target, ir.Var)
    assert isinstance(strs, tuple)

    # Enumerate the strings we're going to be reducing over
    enumerated_conjunctions = []
    index_var, index_type = gen_unique_var(ctx, "idx"), meta_type_to_lqp(types.Int64)
    string_var, string_type = gen_unique_var(ctx, "s"), meta_type_to_lqp(types.String)
    for i, s in enumerate(strs):
        index = to_lqp_value(i, types.Int64)
        string = _translate_term(ctx, s)[0]
        eq_idx = mk_primitive("rel_primitive_eq", [index_var, index])
        eq_string = mk_primitive("rel_primitive_eq", [string_var, string])
        enumerated_conjunctions.append(mk_and([eq_idx, eq_string]))
    body = mk_abstraction([(index_var, index_type), (string_var, string_type)], mk_or(enumerated_conjunctions))

    # Make the function we're reducing
    a_var, a_type = gen_unique_var(ctx, "a"), meta_type_to_lqp(types.String)
    b_var, b_type = gen_unique_var(ctx, "b"), meta_type_to_lqp(types.String)
    curr_var, curr_type = gen_unique_var(ctx, "curr"), meta_type_to_lqp(types.String)
    res_var, res_type = gen_unique_var(ctx, "res"), meta_type_to_lqp(types.String)

    sep, _ = _translate_term(ctx, separator)

    op = mk_abstraction(
        [(a_var, a_type), (b_var, b_type), (curr_var, curr_type)],
        mk_exists([(res_var, res_type)],
            mk_and([
                mk_primitive("rel_primitive_concat", [res_var, b_var, curr_var]),
                mk_primitive("rel_primitive_concat", [a_var, sep, res_var])
            ])
        )
    )

    output_term = _translate_term(ctx, target)[0]

    return lqp_bundle.ir.Reduce(meta=None, op=op, body=body, terms=[output_term])

def _translate_var(ctx: TranslationCtx, term: ir.Var) -> lqp_bundle.ir.Var:
    name = ctx.var_names.get_name_by_id(term.id, term.name)
    return lqp_bundle.ir.Var(name=name, meta=None)



"""Python IR dataclasses that match the old lqp 0.2.x API.

These classes provide compatibility with the old lqp.ir module that was removed in
lqp 0.4.0. They are pure Python dataclasses that can be converted to lqp 0.4.0 protobuf
messages via the ir_to_proto function in proto_conversion.py.
"""
from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from typing import Any


# --- Abstract base classes for IR type hierarchies ---

class RelTerm(ABC):
    """Base class for relation terms (Var, Value, SpecializedValue)."""

class Term(RelTerm, ABC):
    """Base class for terms (Var, Value). All Terms are also RelTerms."""

class Formula(ABC):
    """Base class for formulas (Atom, Conjunction, Disjunction, etc.)."""

class Declaration(ABC):
    """Base class for declarations (Def, FunctionalDependency, Algorithm)."""

class Construct(ABC):
    """Base class for constructs in scripts (Loop, Assign, Upsert, etc.)."""

class Instruction(Construct, ABC):
    """Base class for instructions (Assign, Upsert, MonoidDef, MonusDef, Break). All Instructions are Constructs."""


# --- Source info (used as type hints only) ---

@dataclass(frozen=True)
class SourceInfo:
    meta: Any = None


@dataclass(frozen=True)
class LqpNode:
    meta: Any = None


# --- Type system ---

class TypeName(Enum):
    INT = "INT"
    INT128 = "INT128"
    UINT128 = "UINT128"
    FLOAT = "FLOAT"
    STRING = "STRING"
    DATE = "DATE"
    DATETIME = "DATETIME"
    DECIMAL = "DECIMAL"
    BOOLEAN = "BOOLEAN"
    MISSING = "MISSING"


@dataclass(frozen=True)
class Type:
    type_name: TypeName
    parameters: list = field(default_factory=list)
    meta: Any = None


# --- Relation ID ---

@dataclass(frozen=True, eq=False)
class RelationId:
    """128-bit relation identifier."""

    id: int  # noqa: A003
    meta: Any = field(default=None, compare=False, hash=False)

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, RelationId):
            return self.id == other.id
        return NotImplemented

    def __repr__(self) -> str:
        return f"RelationId(id={self.id!r})"


# --- Variable ---

@dataclass(frozen=True)
class Var(Term):
    name: str
    meta: Any = None


# --- Value types ---

@dataclass(frozen=True)
class Value(Term):
    """Wraps a raw Python value or a typed value wrapper (Int128Value, DateValue, etc.)."""
    value: Any
    meta: Any = None


@dataclass(frozen=True)
class Int128Value:
    value: int
    meta: Any = None


@dataclass(frozen=True)
class UInt128Value:
    value: int
    meta: Any = None


@dataclass(frozen=True)
class DecimalValue:
    precision: int
    scale: int
    value: Any  # decimal.Decimal
    meta: Any = None


@dataclass(frozen=True)
class DateValue:
    value: Any  # datetime.date
    meta: Any = None


@dataclass(frozen=True)
class DateTimeValue:
    value: Any  # datetime.datetime
    meta: Any = None


@dataclass(frozen=True)
class BooleanValue:
    value: bool
    meta: Any = None


@dataclass(frozen=True)
class SpecializedValue(RelTerm):
    """A specialized (typed) value used as a RelTerm in RelAtom/Primitive terms."""
    value: Value
    meta: Any = None


# --- Attributes ---

@dataclass(frozen=True)
class Attribute:
    name: str
    args: list = field(default_factory=list)  # list[Value]
    meta: Any = None


# --- Abstraction ---

@dataclass(frozen=True)
class Abstraction:
    """Lambda abstraction: vars is a list of (Var, Type) tuples, value is a Formula."""
    vars: list  # list[(Var, Type)]
    value: Any  # Formula
    meta: Any = None


# --- Formula types ---

@dataclass(frozen=True)
class Conjunction(Formula):
    args: list  # list[Formula]
    meta: Any = None


@dataclass(frozen=True)
class Disjunction(Formula):
    args: list  # list[Formula]
    meta: Any = None


@dataclass(frozen=True)
class Not(Formula):
    arg: Any  # Formula
    meta: Any = None


@dataclass(frozen=True)
class Exists(Formula):
    body: Abstraction
    meta: Any = None


@dataclass(frozen=True)
class Reduce(Formula):
    op: Abstraction
    body: Abstraction
    terms: list  # list[Term]
    meta: Any = None


@dataclass(frozen=True)
class FFI(Formula):
    name: str
    args: list  # list[Abstraction]
    terms: list  # list[Term]
    meta: Any = None


@dataclass(frozen=True)
class Atom(Formula):
    name: RelationId
    terms: list  # list[Term]
    meta: Any = None


@dataclass(frozen=True)
class Pragma(Formula):
    name: str
    terms: list  # list[Var]
    meta: Any = None


@dataclass(frozen=True)
class Primitive(Formula):
    name: str
    terms: list  # list[RelTerm]
    meta: Any = None


@dataclass(frozen=True)
class RelAtom(Formula):
    name: str
    terms: list  # list[RelTerm]
    meta: Any = None


@dataclass(frozen=True)
class Cast(Formula):
    input: Any   # Term
    result: Any  # Term
    meta: Any = None


# --- Declaration types ---

@dataclass(frozen=True)
class Def(Declaration):
    name: RelationId
    body: Abstraction
    attrs: list = field(default_factory=list)  # list[Attribute]
    meta: Any = None


class Constraint:
    """Base class for constraint-type declarations."""
    pass


@dataclass(frozen=True)
class FunctionalDependency(Declaration, Constraint):
    name: RelationId
    guard: Abstraction
    keys: list  # list[Var]
    values: list  # list[Var]
    meta: Any = None


# --- Algorithm / Script / Loopy constructs ---

@dataclass(frozen=True)
class OrMonoid:
    meta: Any = None


@dataclass(frozen=True)
class SumMonoid:
    type: Any = None  # Type
    meta: Any = None


@dataclass(frozen=True)
class MinMonoid:
    type: Any = None  # Type
    meta: Any = None


@dataclass(frozen=True)
class MaxMonoid:
    type: Any = None  # Type
    meta: Any = None


@dataclass(frozen=True)
class Assign(Instruction):
    name: RelationId
    body: Abstraction
    attrs: list = field(default_factory=list)  # list[Attribute]
    meta: Any = None


@dataclass(frozen=True)
class Upsert(Instruction):
    name: RelationId
    body: Abstraction
    attrs: list = field(default_factory=list)  # list[Attribute]
    value_arity: int = 0
    meta: Any = None


@dataclass(frozen=True)
class MonoidDef(Instruction):
    monoid: Any  # OrMonoid | SumMonoid | MinMonoid | MaxMonoid
    name: Any = None  # RelationId
    body: Any = None  # Abstraction
    attrs: list = field(default_factory=list)  # list[Attribute]
    value_arity: int = 0
    meta: Any = None


@dataclass(frozen=True)
class MonusDef(Instruction):
    monoid: Any  # OrMonoid | SumMonoid | MinMonoid | MaxMonoid
    name: Any = None  # RelationId
    body: Any = None  # Abstraction
    attrs: list = field(default_factory=list)  # list[Attribute]
    value_arity: int = 0
    meta: Any = None


@dataclass(frozen=True)
class Break(Instruction):
    """Break condition in a loopy script."""
    name: RelationId
    body: Abstraction
    attrs: list = field(default_factory=list)  # list[Attribute]
    meta: Any = None


@dataclass(frozen=True)
class Loop(Construct):
    init: list = field(default_factory=list)  # list[Instruction]
    body: Any = None  # Script
    meta: Any = None


@dataclass(frozen=True)
class Script:
    constructs: list = field(default_factory=list)  # list[Construct]
    meta: Any = None


@dataclass(frozen=True)
class Algorithm(Declaration):
    global_: list = field(default_factory=list)  # list[RelationId]
    body: Any = None  # Script
    meta: Any = None


# --- Transaction types ---

@dataclass(frozen=True, eq=False)
class FragmentId:
    """Fragment identifier (bytes)."""

    id: bytes  # noqa: A003
    meta: Any = field(default=None, compare=False, hash=False)

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, FragmentId):
            return self.id == other.id
        return NotImplemented

    def __repr__(self) -> str:
        return f"FragmentId(id={self.id!r})"


@dataclass(frozen=True)
class DebugInfo:
    """Maps RelationId → original name for debugging."""
    id_to_orig_name: dict = field(default_factory=dict)  # RelationId -> str
    meta: Any = None


@dataclass(frozen=True)
class Fragment:
    """A compilation fragment containing declarations."""
    id: FragmentId
    declarations: list  # list[Declaration]
    debug_info: DebugInfo
    meta: Any = None


@dataclass(frozen=True)
class Define:
    fragment: Fragment
    meta: Any = None


@dataclass(frozen=True)
class Undefine:
    fragment_id: FragmentId
    meta: Any = None


@dataclass(frozen=True)
class Write:
    """A write operation. write_type is either Define or Undefine."""
    write_type: Any  # Define | Undefine
    meta: Any = None


@dataclass(frozen=True)
class Output:
    name: str
    relation_id: RelationId
    meta: Any = None


@dataclass(frozen=True)
class ExportCSVColumn:
    column_name: str
    column_data: RelationId
    meta: Any = None


@dataclass(frozen=True)
class ExportCSVConfig:
    path: str
    data_columns: list = field(default_factory=list)  # list[ExportCSVColumn]
    compression: str = ""
    partition_size: int = 0
    syntax_header_row: bool = True
    syntax_missing_string: str = ""
    syntax_delim: str = ","
    syntax_quotechar: str = '"'
    syntax_escapechar: str = ""
    meta: Any = None


@dataclass(frozen=True)
class Export:
    """CSV export read operation. config is an ExportCSVConfig."""
    config: ExportCSVConfig
    meta: Any = None




# --- Inline CSV data (data-in-LQP) ---

@dataclass(frozen=True)
class CSVLocator:
    paths: list = field(default_factory=list)
    inline_data: bytes = b""
    meta: Any = None


@dataclass(frozen=True)
class CSVConfig:
    header_row: int = 0
    skip: int = 0
    new_line: str = "\n"
    delimiter: str = ","
    quotechar: str = '"'
    escapechar: str = '"'
    comment: str = ""
    missing_strings: list = field(default_factory=list)
    decimal_separator: str = "."
    encoding: str = "utf-8"
    compression: str = "none"
    meta: Any = None


@dataclass(frozen=True)
class GNFColumn:
    column_name: str = ""
    target_id: RelationId = None  # type: ignore[assignment]
    types: list = field(default_factory=list)  # list[Type]
    meta: Any = None


@dataclass(frozen=True)
class CSVData(Declaration):
    """Inline CSV data declaration."""
    locator: CSVLocator = None  # type: ignore[assignment]
    config: CSVConfig = None  # type: ignore[assignment]
    columns: list = field(default_factory=list)  # list[GNFColumn]
    asof: str = ""
    meta: Any = None

@dataclass(frozen=True)
class Read:
    """A read operation. read_type is either Output or Export."""
    read_type: Any  # Output | Export
    meta: Any = None


@dataclass(frozen=True)
class Sync:
    fragments: list = field(default_factory=list)  # list[FragmentId]
    meta: Any = None


@dataclass(frozen=True)
class Epoch:
    writes: list = field(default_factory=list)   # list[Write]
    reads: list = field(default_factory=list)    # list[Read]
    meta: Any = None


@dataclass(frozen=True)
class IVMConfig:
    level: Any = None  # MaintenanceLevel value
    meta: Any = None


@dataclass(frozen=True)
class Configure:
    semantics_version: int = 0
    ivm_config: Any = None  # IVMConfig | None
    meta: Any = None


@dataclass(frozen=True)
class Transaction:
    epochs: list = field(default_factory=list)  # list[Epoch]
    configure: Any = None  # Configure
    sync: Any = None       # Sync | None
    meta: Any = None


class MaintenanceLevel(Enum):
    UNSPECIFIED = 0
    OFF = 1
    AUTO = 2
    ALL = 3



"""
Lifts Data nodes to top-level GNF Logicals with fresh per-column predicates.

This pass transforms Data nodes that appear nested within Logicals into a set of
top-level GNF (Graph Normal Form) Logicals — one per column plus one for
the row ID.  Each GNF Logical contains the annotated Data node and a derive
statement for the corresponding 1- or 2-arity column predicate.  The original
Data node is replaced with a collection of Lookups to those column predicates.

Example transformation:

Before:
  root: Logical
    ...

    Logical
      Data{df}(row_id, col_1, col_2, col_3)
      predicate(col_1, col_3)

After:
  root: Logical
    Logical @lifted_data
      Data{df}(row_id, col_1, col_2, col_3)
      Logical @gnf_data("__ROW_ID__")
        derive _data_1___row_id__(row_id)
      Logical @gnf_data("col_1")
        derive _data_1_col_1(row_id, col_1)
      Logical @gnf_data("col_2")
        derive _data_1_col_2(row_id, col_2)
      Logical @gnf_data("col_3")
        derive _data_1_col_3(row_id, col_3)
    ...
    Logical
      Logical
        _data_1___row_id__(row_id)
        _data_1_col_1(row_id, col_1)
        _data_1_col_3(row_id, col_3)
      predicate(col_1, col_3)


Temporarily, for backward compatibility, we transform single column Data nodes with less
than 5 values (rows) into a disjunction of equality predicates.

Before:
  root:
    Logical
      ...
      Logical
        Data{[1, 2]}(x) <- small df, less than 5 values
        predicate(x)

After:
  root:
    Logical
      ...
      Logical
        Union
          Logical
            x = 1
          Logical
            x = 2
        predicate(x)

"""

from relationalai.semantics.metamodel.compiler import Pass
from relationalai.semantics.metamodel import ir, builtins as rel_builtins, factory as f, visitor
from relationalai.semantics.lqp.builtins import lifted_data_annotation, gnf_data_annotation, ROW_ID_COLUMN
import hashlib
import pandas as pd


class LiftData(Pass):
    def __init__(self):
        # Persistent counter across compilations so that data node names are globally unique
        # (the same LiftData instance is reused for model and query compilations).
        self.lifted_data_node_count = 0

    def rewrite(self, model: ir.Model, options: dict = {}) -> ir.Model:  # noqa: ARG002
        lifter = DataLifter(lifted_data_node_count=self.lifted_data_node_count)
        result = lifter.walk(model)
        self.lifted_data_node_count = lifter.lifted_data_node_count
        return result


class DataLifter(visitor.Rewriter):
    _EXPAND_MAX_ROWS = 5

    def __init__(self, lifted_data_node_count: int = 0):
        # GNF path state
        self.gnf_lifted_data: list[ir.Logical] = []   # one lifted data Logical (annotated @lifted_data) per unique Data
        self.gnf_relations: list[ir.Relation] = []   # all derived relations

        # Cache: Data node content hash -> [(ROW_ID_COLUMN, _data_k___row_id__), ("col_1", _data_k_col_1), ...]
        self.gnf_cache: dict[str, list[tuple[str, ir.Relation]]] = {}
        self.lifted_data_node_count = lifted_data_node_count

        super().__init__()

    def handle_model(self, model: ir.Model, parent: None) -> ir.Model:
        model = super().handle_model(model, parent)

        if not self.gnf_lifted_data:
            return model

        assert isinstance(model.root, ir.Logical)
        root = model.root
        return ir.Model(
            model.engines,
            model.relations | self.gnf_relations,
            model.types,
            ir.Logical(
                root.engine,
                root.hoisted,
                tuple(self.gnf_lifted_data) + root.body,
                root.annotations,
            ),
            model.annotations,
        )

    def handle_data(self, node: ir.Data, parent: ir.Node) -> ir.Logical:
        """Replace a Data node with a Logical containing Lookups (GNF) or a Union (backward compat)."""
        if self.should_expand(node):
            return self.expand_disjunction(node)
        else:
            return self.gnf_wrapper(node)

    # ------------------------------------------------------------------ #
    # GNF path                                                             #
    # ------------------------------------------------------------------ #

    def gnf_wrapper(self, node: ir.Data) -> ir.Logical:
        """Return a wrapper Logical whose body is the Lookup sequence for node."""
        return f.logical(self.gnf_lookups(node), list(node.vars))

    def gnf_lookups(self, node: ir.Data) -> list[ir.Task]:
        """Return [Lookup(row_id_rel, [row_id]), Lookup(col_rel, [row_id, col]), …]."""
        row_id_var = node.vars[0]
        signature = self.get_data_gnf_signature(node)
        # First entry is the 1-arity __ROW_ID__ predicate
        lookups: list[ir.Task] = [f.lookup(signature[0][1], [row_id_var])]
        # Remaining entries are 2-arity column predicates: (row_id, col_var)
        col_vars = [node.vars[i] for i in range(1, len(node.vars))]
        for (_, rel), var in zip(signature[1:], col_vars):
            lookups.append(f.lookup(rel, [row_id_var, var]))
        return lookups

    def get_data_gnf_signature(self, node: ir.Data) -> list[tuple[str, ir.Relation]]:
        """Return cached (name, relation) pairs for node, lifting it on first call.

        Returns:
            A list where index 0 is (ROW_ID_COLUMN, row_id_relation) — 1-arity,
            and indices 1+ are (col_name, col_relation) — 2-arity each.
        """
        cache_key = self.data_content_hash(node)
        if cache_key not in self.gnf_cache:
            self.gnf_cache[cache_key] = self.lift_data_gnf(node)
        return self.gnf_cache[cache_key]

    def lift_data_gnf(self, node: ir.Data) -> list[tuple[str, ir.Relation]]:
        """Lift a Data node: build per-column relations, register the lifted data Logical."""
        self.lifted_data_node_count += 1
        signature, gnf_children = self.mk_gnf_signature(node, self.lifted_data_node_count)
        self.gnf_lifted_data.append(f.logical(
            [node] + gnf_children,
            annos=[lifted_data_annotation()],
        ))
        return signature

    def mk_gnf_signature(
        self, node: ir.Data, n: int
    ) -> tuple[list[tuple[str, ir.Relation]], list[ir.Task]]:
        """Build per-column relations and @gnf_data derive nodes for node.

        Returns:
            (signature, gnf_children) where signature is the (name, relation) list
            and gnf_children are the @gnf_data derive nodes.
        """
        result: list[tuple[str, ir.Relation]] = []
        gnf_children: list[ir.Task] = []
        row_id_var = node.vars[0]

        # 1-arity row-ID predicate
        row_id_rel = f.relation(
            f"_data_{n}_{ROW_ID_COLUMN}",
            [f.field(row_id_var.name, row_id_var.type)],
        )
        self.gnf_relations.append(row_id_rel)
        result.append((ROW_ID_COLUMN, row_id_rel))
        gnf_children.append(f.logical(
            [f.derive(row_id_rel, [row_id_var])],
            annos=[gnf_data_annotation(ROW_ID_COLUMN)],
        ))

        # 2-arity per-column predicates
        col_vars = [node.vars[i] for i in range(1, len(node.vars))]
        for col_name, var in zip(node.data.columns, col_vars):
            col_rel = f.relation(
                f"_data_{n}_{col_name}",
                [
                    f.field(row_id_var.name, row_id_var.type),
                    f.field(var.name, var.type),
                ],
            )
            self.gnf_relations.append(col_rel)
            result.append((col_name, col_rel))
            gnf_children.append(f.logical(
                [f.derive(col_rel, [row_id_var, var])],
                annos=[gnf_data_annotation(col_name)],
            ))

        return result, gnf_children

    # ------------------------------------------------------------------ #
    # Backward compatibility path (single-column, < 5 rows)              #
    # These two methods form a self-contained code path that can be      #
    # deleted once the GNF path is fully adopted.                        #
    # ------------------------------------------------------------------ #

    @staticmethod
    def should_expand(node: ir.Data) -> bool:
        """True for single-column (excluding row_id) Data nodes with fewer than _EXPAND_MAX_ROWS rows."""
        return len(node.data.columns) == 1 and 0 < len(node.data) < DataLifter._EXPAND_MAX_ROWS

    @staticmethod
    def expand_disjunction(node: ir.Data) -> ir.Logical:
        """Expand a small single-column Data into a wrapper Logical containing a Union.

        The Union has one branch per row; each branch equates the row-id var and the
        column var to their literal values from the DataFrame.
        """
        row_id_var = node.vars[0]
        col_var = node.vars[1]
        branches: list[ir.Task] = []
        for row in node:  # itertuples(index=True): Pandas(Index=idx, col=val)
            idx = row[0]
            col_val = row[1]
            branch = f.logical(
                [
                    f.lookup(rel_builtins.eq, [row_id_var, f.literal(DataLifter.to_python(idx), type=row_id_var.type)]),
                    f.lookup(rel_builtins.eq, [col_var, f.literal(DataLifter.to_python(col_val), type=col_var.type)]),
                ],
                hoisted=list(node.vars),
            )
            branches.append(branch)
        union = f.union(branches, hoisted=list(node.vars))
        return f.logical([union])

    @staticmethod
    def to_python(val):
        """Convert a numpy scalar to a plain Python value."""
        return val.item() if hasattr(val, "item") else val

    # ------------------------------------------------------------------ #
    # Shared helpers                                                       #
    # ------------------------------------------------------------------ #

    @staticmethod
    def data_content_hash(node: ir.Data) -> str:
        """Content-based cache key incorporating column names, var types, and row data."""
        col_names = ",".join(str(c) for c in node.data.columns)
        col_types = ",".join(str(v.type) for v in node.vars)
        row_bytes = bytes(pd.util.hash_pandas_object(node.data, index=True).values)
        return hashlib.sha256(f"{col_names}|{col_types}|".encode() + row_bytes).hexdigest()



"""Solver engine lifecycle and job execution.

Provides the Solver and Provider classes for managing solver engines
and executing optimization jobs on RelationalAI's solver infrastructure.
"""

from __future__ import annotations

import json
import logging
import textwrap
from dataclasses import dataclass
from typing import Any, Optional, cast

import relationalai
from relationalai import debugging
from relationalai.clients.direct_access_client import DirectAccessClient
from relationalai.clients.resources.snowflake import (
    APP_NAME,
    DirectAccessResources,
    Resources as SnowflakeResources,
)
from relationalai.clients.util import poll_with_specified_overhead
from relationalai.errors import QueryTimeoutExceededException, ResponseStatusException
from relationalai.tools.cli_controls import Spinner
from relationalai.tools.constants import DEFAULT_QUERY_TIMEOUT_MINS

logger = logging.getLogger(__name__)

ENGINE_TYPE_SOLVER = "SOLVER"
# TODO (dba) The ERP still uses `worker` instead of `engine`. Change
# this once we fix this in the ERP.
WORKER_ERRORS = [
    "worker is suspended",
    "create/resume",
    "worker not found",
    "no workers found",
    "worker was deleted",
]
ENGINE_ERRORS = [
    "engine is suspended",
    "create/resume",
    "engine not found",
    "no engines found",
    "engine was deleted",
]
ENGINE_NOT_READY_MSGS = [
    "worker is in pending",
    "worker is provisioning",
    "worker is not ready to accept jobs",
    "resume in progress",
]

JOB_ABORT_REASON_TIMEOUT = "job timeout"


@dataclass
class PollingState:
    job_id: str
    continuation_token: str
    is_done: bool
    log_to_console: bool


class Solver:
    def __init__(
        self,
        solver_name: str,
        engine_name: str | None = None,
        engine_size: str | None = None,
        auto_suspend_mins: int | None = None,
        resources: SnowflakeResources | None = None,
    ):
        self.provider = Provider(resources=resources)
        self.solver_name = solver_name.lower()

        self.rai_config = self.provider.resources.config
        settings: dict[str, Any] = {}
        if "experimental" in self.rai_config:
            exp_config = self.rai_config.get("experimental", {})
            if isinstance(exp_config, dict):
                if "solvers" in exp_config:
                    settings = exp_config["solvers"].copy()

        # Engine configuration fields are not necessary for the solver
        # settings so we `pop` them from the settings object. Default
        # size and auto_suspend_mins are set in the `Provider` methods.
        engine_name = engine_name or settings.pop("engine", None)
        if not engine_name:
            engine_name = self.provider.resources.get_user_based_engine_name()
        self.engine_name = engine_name

        self.engine_size = engine_size or settings.pop("engine_size", None)
        self.engine_auto_suspend_mins = auto_suspend_mins or settings.pop(
            "auto_suspend_mins", None
        )

        # Set default CSV store setting if not already configured
        if "store" not in settings:
            settings["store"] = {}
        if "csv" not in settings["store"]:
            settings["store"]["csv"] = {}
        if "enabled" not in settings["store"]["csv"]:
            settings["store"]["csv"]["enabled"] = True

        # The settings are used when creating a solver engine, they
        # may configure each individual solver.
        self.engine_settings = settings

        # Optimistically set the engine object to a `READY` engine to
        # avoid checking the engine status on each execution.
        self.engine: Optional[dict[str, Any]] = {"name": engine_name, "state": "READY"}

        return None

    # --------------------------------------------------
    # Helper
    # --------------------------------------------------
    def _auto_create_solver_async(self):
        name = self.engine_name
        auto_suspend_mins = self.engine_auto_suspend_mins
        size = self.engine_size
        settings = self.engine_settings
        with Spinner(
            "Checking solver status",
            leading_newline=True,
        ) as spinner:
            engine = None
            engines = [e for e in self.provider.list_solvers() if e["name"] == name]
            if len(engines) > 1:
                raise RuntimeError(
                    f"Expected at most 1 engine named '{name}', but found {len(engines)}."
                )
            if len(engines) != 0:
                engine = engines[0]

            if engine:
                # TODO (dba) Logic engines support altering the
                # auto_suspend_mins setting. Currently, we don't have
                # this capability for solver engines, so users need to
                # recreate or use another engine. For both the size
                # and Gurobi configuration the user anyways has to
                # create a new one as this configuration must happen
                # when the engine is created.
                settings_cannot_be_altered_msg = "The configuration of a solver engine happens when the engine is created and _cannot_ be changed. You either need to specify a new engine name or delete the current engine.\n\nSee `solvers.Provider().delete_solver()`."
                # Make sure that the solver requested is enabled
                # on the engine.
                if self.solver_name not in engine["solvers"]:
                    raise Exception(
                        f"Solver `{self.solver_name}` is not enabled on `{name}`.\n\n"
                        + settings_cannot_be_altered_msg
                    )

                # Make sure size and auto_suspend_mins settings match
                # what the user requests.
                if size is not None and size != engine["size"]:
                    raise Exception(
                        f"Engine `{name}` has size setting of `{engine['size']}` but size is requested to be `{size}`.\n\n"
                        + settings_cannot_be_altered_msg
                    )

                if (
                    auto_suspend_mins is not None
                    and auto_suspend_mins != engine["auto_suspend_mins"]
                ):
                    raise Exception(
                        f"Engine `{name}` has auto_suspend_mins setting of `{engine['auto_suspend_mins']}` but auto_suspend_mins is requested to be `{auto_suspend_mins}`.\n\n"
                        + settings_cannot_be_altered_msg
                    )

                if engine["state"] == "PENDING":
                    spinner.update_messages(
                        {
                            "finished_message": f"Solver {name} is starting",
                        }
                    )
                    pass
                elif engine["state"] == "SUSPENDED":
                    spinner.update_messages(
                        {
                            "finished_message": f"Resuming solver {name}",
                        }
                    )
                    self.provider.resume_solver_async(name)
                elif engine["state"] == "READY":
                    spinner.update_messages(
                        {
                            "finished_message": f"Solver {name} is ready",
                        }
                    )
                    pass
                else:
                    spinner.update_messages(
                        {
                            "message": f"Restarting solver {name}",
                        }
                    )
                    self.provider.delete_solver(name)
                    engine = None
            if not engine:
                # Validate Gurobi config.
                if self.solver_name == "gurobi":
                    is_gurobi_configured = False
                    gurobi_config = settings.get("gurobi", {})
                    if all(
                        k in gurobi_config
                        for k in ["license_secret_name", "external_access_integration"]
                    ):
                        is_gurobi_configured = True
                    if not is_gurobi_configured:
                        raise Exception(
                            "Gurobi is not properly configured. You need to provide both `license_secret_name` and `external_access_integration` in its configuration, see https://docs.relational.ai/build/reasoners/prescriptive/solver-backends/gurobi/#usage"
                        )
                self.provider.create_solver_async(
                    name,
                    settings=settings,
                    size=size,
                    auto_suspend_mins=auto_suspend_mins,
                )
                engine = self.provider.get_solver(name)
                spinner.update_messages(
                    {
                        "finished_message": f"Starting solver {name}...",
                    }
                )

            self.engine = engine

    def _exec_job(
        self, payload, log_to_console=True, query_timeout_mins: Optional[int] = None
    ):
        if self.engine is None:
            raise Exception("Engine not initialized.")

        with debugging.span("job") as job_span:
            # Retry logic. If creating a job fails with an engine
            # related error we will create/resume/... the engine and
            # retry.
            try:
                job_id = self.provider.create_job_async(
                    self.engine["name"], payload, query_timeout_mins=query_timeout_mins
                )
            except Exception as e:
                err_message = str(e).lower()
                if isinstance(e, ResponseStatusException):
                    try:
                        err_message = e.response.json().get("message", "")
                    except (ValueError, AttributeError):
                        err_message = e.response.text if hasattr(e.response, 'text') else str(e)
                if any(
                    kw in err_message.lower()
                    for kw in ENGINE_ERRORS + WORKER_ERRORS + ENGINE_NOT_READY_MSGS
                ):
                    self._auto_create_solver_async()
                    # Wait until the engine is ready.
                    poll_with_specified_overhead(lambda: self._is_solver_ready(), 0.1)
                    job_id = self.provider.create_job_async(
                        self.engine["name"],
                        payload,
                        query_timeout_mins=query_timeout_mins,
                    )
                else:
                    raise

            job_span["job_id"] = job_id
            debugging.event(
                "job_created",
                job_span,
                job_id=job_id,
                engine_name=self.engine["name"],
                job_type=ENGINE_TYPE_SOLVER,
            )
            if not isinstance(job_id, str):
                job_id = ""
            polling_state = PollingState(job_id, "", False, log_to_console)

            try:
                with debugging.span("wait", job_id=job_id):
                    poll_with_specified_overhead(
                        lambda: self._check_job_status(polling_state), 0.1
                    )
            except KeyboardInterrupt as e:
                print(f"Canceling job {job_id}")
                self.provider.cancel_job(job_id)
                raise e

            return job_id

    def _is_solver_ready(self):
        if self.engine is None:
            raise Exception("Engine not initialized.")

        result = self.provider.get_solver(self.engine["name"])

        if result is None:
            raise Exception("No engine available.")

        self.engine = result
        state = result["state"]
        if state != "READY" and state != "PENDING":
            # Might have suspended or otherwise gone. Recreate.
            self._auto_create_solver_async()
        return state == "READY"

    def _check_job_status(self, state):
        response = self.provider.get_job(state.job_id)
        if not response:
            raise RuntimeError(f"No results from get_job('{state.job_id}')")

        status: str = response["state"]
        abort_reason = response.get("abort_reason", "")

        self._print_solver_logs(state)
        if status == "CANCELED" and abort_reason == JOB_ABORT_REASON_TIMEOUT:
            config_file_path = getattr(self.rai_config, "file_path", None)
            timeout_ms = int(response.get("timeout_ms", 0))
            timeout_mins = (
                timeout_ms // 60000
                if timeout_ms > 0
                else int(
                    self.rai_config.get(
                        "query_timeout_mins", DEFAULT_QUERY_TIMEOUT_MINS
                    )
                    or DEFAULT_QUERY_TIMEOUT_MINS
                )
            )
            raise QueryTimeoutExceededException(
                timeout_mins=timeout_mins,
                query_id=state.job_id,
                config_file_path=config_file_path,
            )

        if status == "FAILED":
            raise RuntimeError(
                f"Solver job '{state.job_id}' failed."
            )
        if status == "CANCELED":
            raise RuntimeError(
                f"Solver job '{state.job_id}' was canceled. "
                f"Reason: {abort_reason or 'unknown'}"
            )

        return status == "COMPLETED"

    def _print_solver_logs(self, state: PollingState):
        if state.is_done:
            return

        resp = self.provider.get_job_events(state.job_id, state.continuation_token)

        # Print solver logs to stdout.
        for event in resp["events"]:
            if event["type"] == "LogMessage":
                if state.log_to_console:
                    print(event["event"]["message"])
            else:
                continue

        state.continuation_token = resp["continuation_token"]
        if state.continuation_token == "":
            state.is_done = True


# --------------------------------------------------
# Provider
# --------------------------------------------------
#
# TODO (dba) We use an experimental and unified engine API for
# solvers. Once it is no longer experimental we can remove the
# provider here and use the normal PyRel provider.


class Provider:
    def __init__(self, resources=None):
        if not resources:
            resources = relationalai.Resources()
        if not isinstance(resources, SnowflakeResources):
            raise Exception("Solvers are only supported on SPCS.")

        # Type narrowing: resources is confirmed to be SnowflakeResources
        self.resources: SnowflakeResources = cast(SnowflakeResources, resources)
        self.direct_access_client: Optional[DirectAccessClient] = None

        if isinstance(self.resources, DirectAccessResources):
            self.direct_access_client = self.resources.direct_access_client

    def create_solver(
        self,
        name: str,
        size: str | None = None,
        settings: dict | None = None,
        auto_suspend_mins: int | None = None,
    ):
        if size is None:
            size = "HIGHMEM_X64_S"
        if settings is None:
            settings = {}
        engine_config: dict[str, Any] = {"settings": settings}
        if auto_suspend_mins is not None:
            engine_config["auto_suspend_mins"] = auto_suspend_mins
        self.resources._exec_sql(
            f"CALL {APP_NAME}.experimental.create_engine('{ENGINE_TYPE_SOLVER}', '{name}', '{size}', {engine_config});",
            None,
        )

    def create_solver_async(
        self,
        name: str,
        size: str | None = None,
        settings: dict | None = None,
        auto_suspend_mins: int | None = None,
    ):
        if size is None:
            size = "HIGHMEM_X64_S"

        if self.direct_access_client is not None:
            payload: dict[str, Any] = {
                "name": name,
                "settings": settings,
            }
            if auto_suspend_mins is not None:
                payload["auto_suspend_mins"] = auto_suspend_mins
            if size is not None:
                payload["size"] = size
            response = self.direct_access_client.request(
                "create_engine",
                payload=payload,
                path_params={"engine_type": "solver"},
            )
            if response.status_code != 200:
                raise ResponseStatusException(
                    f"Failed to create engine {name} with size {size}.", response
                )
        else:
            engine_config: dict[str, Any] = {}
            if settings is not None:
                engine_config["settings"] = settings
            if auto_suspend_mins is not None:
                engine_config["auto_suspend_mins"] = auto_suspend_mins
            self.resources._exec_sql(
                f"CALL {APP_NAME}.experimental.create_engine_async('{ENGINE_TYPE_SOLVER}', '{name}', '{size}', {engine_config});",
                None,
            )

    def delete_solver(self, name: str):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "delete_engine",
                path_params={"engine_type": ENGINE_TYPE_SOLVER, "engine_name": name},
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to delete engine.", response)
            return None
        else:
            self.resources._exec_sql(
                f"CALL {APP_NAME}.experimental.delete_engine('{ENGINE_TYPE_SOLVER}', '{name}');",
                None,
            )

    def resume_solver_async(self, name: str):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "resume_engine",
                path_params={"engine_type": ENGINE_TYPE_SOLVER, "engine_name": name},
            )
            if response.status_code != 200:
                # If the engine is already resuming, treat as a no-op.
                try:
                    msg = response.json().get("message", "")
                except (ValueError, AttributeError):
                    msg = response.text
                if "resume in progress" in msg.lower():
                    logger.debug(
                        "Engine %s resume already in progress, treating as no-op.", name
                    )
                    return None
                raise ResponseStatusException("Failed to resume engine.", response)
            return None
        else:
            try:
                self.resources._exec_sql(
                    f"CALL {APP_NAME}.experimental.resume_engine_async('{ENGINE_TYPE_SOLVER}', '{name}');",
                    None,
                )
            except Exception as e:
                # If the engine is already resuming, treat as a no-op.
                if "resume in progress" in str(e).lower():
                    logger.debug(
                        "Engine %s resume already in progress, treating as no-op.", name
                    )
                    return None
                raise
            return None

    def get_solver(self, name: str):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "get_engine",
                path_params={"engine_type": ENGINE_TYPE_SOLVER, "engine_name": name},
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to get engine.", response)
            solver = response.json()
            if not solver:
                return None
            solver_state = {
                "name": solver["name"],
                "id": solver["id"],
                "size": solver["size"],
                "state": solver["status"],  # callers are expecting 'state'
                "created_by": solver["created_by"],
                "created_on": solver["created_on"],
                "updated_on": solver["updated_on"],
                "version": solver["version"],
                "auto_suspend": solver["auto_suspend_mins"],
                "suspends_at": solver["suspends_at"],
                "solvers": []
                if solver["settings"] == ""
                else [
                    k
                    for (k, v) in json.loads(solver["settings"]).items()
                    if isinstance(v, dict) and v.get("enabled", False)
                ],
            }
            return solver_state
        else:
            results = self.resources._exec_sql(
                f"CALL {APP_NAME}.experimental.get_engine('{ENGINE_TYPE_SOLVER}', '{name}');",
                None,
            )
            return solver_list_to_dicts(results)[0]

    def list_solvers(self, state: str | None = None):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request("list_engines")
            if response.status_code != 200:
                raise ResponseStatusException("Failed to list engines.", response)
            response_content = response.json()
            if not response_content:
                return []
            engines = [
                {
                    "name": engine["name"],
                    "id": engine["id"],
                    "size": engine["size"],
                    "state": engine["status"],  # callers are expecting 'state'
                    "created_by": engine["created_by"],
                    "created_on": engine["created_on"],
                    "updated_on": engine["updated_on"],
                    "auto_suspend_mins": engine["auto_suspend_mins"],
                    "solvers": []
                    if engine["settings"] == ""
                    else [
                        k
                        for (k, v) in json.loads(engine["settings"]).items()
                        if isinstance(v, dict) and v.get("enabled", False)
                    ],
                }
                for engine in response_content.get("engines", [])
                if (state is None or engine.get("status") == state)
                and (engine.get("type") == ENGINE_TYPE_SOLVER)
            ]
            return sorted(engines, key=lambda x: x["name"])
        else:
            where_clause = f"WHERE TYPE='{ENGINE_TYPE_SOLVER}'"
            where_clause = (
                f"{where_clause} AND STATUS = '{state.upper()}'"
                if state
                else where_clause
            )
            statement = f"SELECT NAME,ID,SIZE,STATUS,CREATED_BY,CREATED_ON,UPDATED_ON,AUTO_SUSPEND_MINS,SETTINGS FROM {APP_NAME}.experimental.engines {where_clause};"
            results = self.resources._exec_sql(statement, None)
            return solver_list_to_dicts(results)

    # --------------------------------------------------
    # Job API
    # --------------------------------------------------

    def create_job_async(
        self, engine_name, payload, query_timeout_mins: Optional[int] = None
    ):
        payload_json = json.dumps(payload)

        if (
            query_timeout_mins is None
            and (
                timeout_value := self.resources.config.get(
                    "query_timeout_mins", DEFAULT_QUERY_TIMEOUT_MINS
                )
            )
            is not None
        ):
            query_timeout_mins = int(timeout_value)

        if self.direct_access_client is not None:
            job = {
                "job_type": ENGINE_TYPE_SOLVER,
                "worker_name": engine_name,
                "timeout_mins": query_timeout_mins,
                "payload": payload_json,
            }
            response = self.direct_access_client.request(
                "create_job",
                payload=job,
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to create job.", response)
            response_content = response.json()
            return response_content["id"]
        else:
            if query_timeout_mins is not None:
                sql_string = textwrap.dedent(f"""
                CALL {APP_NAME}.experimental.exec_job_async('{ENGINE_TYPE_SOLVER}', '{engine_name}', '{payload_json}', null, {query_timeout_mins})
                """)
            else:
                sql_string = textwrap.dedent(f"""
                CALL {APP_NAME}.experimental.exec_job_async('{ENGINE_TYPE_SOLVER}', '{engine_name}', '{payload_json}')
                """)
            res = self.resources._exec_sql(sql_string, None)
            return res[0]["ID"]

    def list_jobs(self, state=None, limit=None):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request("list_jobs")
            if response.status_code != 200:
                raise ResponseStatusException("Failed to list jobs.", response)
            response_content = response.json()
            if not response_content:
                return []
            jobs = [
                {
                    "id": job["id"],
                    "state": job["state"],
                    "created_by": job["created_by"],
                    "created_on": job["created_on"],
                    "finished_at": job.get("finished_at", None),
                    "duration": job["duration"] if "duration" in job else 0,
                    "solver": json.loads(job["payload"]).get("solver", ""),
                    "engine": job.get("engine_name", job["worker_name"]),
                }
                for job in response_content.get("jobs", [])
                if state is None or job.get("state") == state
            ]
            return sorted(jobs, key=lambda x: x["created_on"], reverse=True)
        else:
            state_clause = f"AND STATE = '{state.upper()}'" if state else ""
            limit_clause = f"LIMIT {limit}" if limit else ""
            results = self.resources._exec_sql(
                f"SELECT ID,STATE,CREATED_BY,CREATED_ON,FINISHED_AT,DURATION,PAYLOAD,ENGINE_NAME FROM {APP_NAME}.experimental.jobs where type='{ENGINE_TYPE_SOLVER}' {state_clause} ORDER BY created_on DESC {limit_clause};",
                None,
            )
            return job_list_to_dicts(results)

    def get_job(self, id: str):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "get_job", path_params={"job_type": ENGINE_TYPE_SOLVER, "job_id": id}
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to get job.", response)
            response_content = response.json()
            return response_content["job"]
        else:
            results = self.resources._exec_sql(
                f"CALL {APP_NAME}.experimental.get_job('{ENGINE_TYPE_SOLVER}', '{id}');",
                None,
            )
            return job_list_to_dicts(results)[0]

    def get_job_events(self, job_id: str, continuation_token: str = ""):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "get_job_events",
                path_params={
                    "job_type": ENGINE_TYPE_SOLVER,
                    "job_id": job_id,
                    "stream_name": "progress",
                },
                query_params={"continuation_token": continuation_token},
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to get job events.", response)
            response_content = response.json()
            if not response_content:
                return {"events": [], "continuation_token": None}
            return response_content
        else:
            results = self.resources._exec_sql(
                f"SELECT {APP_NAME}.experimental.get_job_events('{ENGINE_TYPE_SOLVER}', '{job_id}', '{continuation_token}');",
                None,
            )
            if not results:
                return {"events": [], "continuation_token": None}
            row = results[0][0]
            if not isinstance(row, str):
                row = ""
            return json.loads(row)

    def cancel_job(self, id: str):
        if self.direct_access_client is not None:
            response = self.direct_access_client.request(
                "cancel_job", path_params={"job_type": ENGINE_TYPE_SOLVER, "job_id": id}
            )
            if response.status_code != 200:
                raise ResponseStatusException("Failed to cancel job.", response)
            return None
        else:
            self.resources._exec_sql(
                f"CALL {APP_NAME}.experimental.cancel_job('{ENGINE_TYPE_SOLVER}', '{id}');",
                None,
            )
            return None


def solver_list_to_dicts(results):
    if not results:
        return []
    return [
        {
            "name": row["NAME"],
            "id": row["ID"],
            "size": row["SIZE"],
            "state": row["STATUS"],  # callers are expecting 'state'
            "created_by": row["CREATED_BY"],
            "created_on": row["CREATED_ON"],
            "updated_on": row["UPDATED_ON"],
            "auto_suspend_mins": row["AUTO_SUSPEND_MINS"],
            "solvers": []
            if row["SETTINGS"] == ""
            else [
                k
                for (k, v) in json.loads(row["SETTINGS"]).items()
                if isinstance(v, dict) and v.get("enabled", False)
            ],
        }
        for row in results
    ]


def job_list_to_dicts(results):
    if not results:
        return []
    return [
        {
            "id": row["ID"],
            "state": row["STATE"],
            "created_by": row["CREATED_BY"],
            "created_on": row["CREATED_ON"],
            "finished_at": row["FINISHED_AT"],
            "duration": row["DURATION"] if "DURATION" in row else 0,
            "solver": json.loads(row["PAYLOAD"])["solver"],
            "engine": row["ENGINE_NAME"],
            "abort_reason": row["ABORT_REASON"] if "ABORT_REASON" in row else "",
            "timeout_ms": row["TIMEOUT_MS"] if "TIMEOUT_MS" in row else 0,
        }
        for row in results
    ]



from __future__ import annotations
from collections import defaultdict
from datetime import datetime, timezone
import atexit
import re
from urllib.parse import urlparse

from pandas import DataFrame
from typing import Any, Optional, Literal, TYPE_CHECKING
from dataclasses import replace
from snowflake.snowpark import Session

from relationalai import debugging
from relationalai.errors import NonDefaultLQPSemanticsVersionWarning
from relationalai.semantics.lqp import result_helpers, export_rewriter, utils
from relationalai.semantics.lqp.utils import ExportInfo, extract_export_fqn
from relationalai.semantics.metamodel import ir, factory as f, executor as e
from relationalai.semantics.metamodel.visitor import collect_by_type

if TYPE_CHECKING:
    from relationalai.semantics.internal.internal import Model as InternalModel
from relationalai.semantics.lqp.compiler import Compiler
from relationalai.semantics.lqp.intrinsics import mk_intrinsic_datetime_now
from relationalai.semantics.lqp.constructors import mk_transaction
from relationalai.semantics.lqp.types import lqp_type_to_sql
from lqp.proto_validator import validate_proto
from relationalai.semantics.lqp import lqp_bundle
from relationalai.clients.config import Config
from relationalai.clients.resources.snowflake import APP_NAME, create_resources_instance
from relationalai.clients.types import TransactionAsyncResponse
from relationalai.clients.util import IdentityParser, escape_for_f_string
from relationalai.tools.constants import QUERY_ATTRIBUTES_HEADER
from relationalai.tools.query_utils import prepare_metadata_for_headers

if TYPE_CHECKING:
    from relationalai.semantics.snowflake import Table

# Whenever the logic engine introduces a breaking change in behaviour, we bump this version
# once the client is ready to handle it.
#
# [2026-01-09] bumping to 1 to opt-into hard validation errors from the engine
DEFAULT_LQP_SEMANTICS_VERSION = "1"

class LQPExecutor(e.Executor):
    """Executes LQP using the RAI client."""

    def __init__(
        self,
        database: str,
        dry_run: bool = False,
        keep_model: bool = True,
        wide_outputs: bool = False,
        connection: Session | None = None,
        config: Config | None = None,
        # In order to facilitate snapshot testing, we allow overriding intrinsic definitions
        # like the current time, which would otherwise change between runs.
        intrinsic_overrides: dict = {},
    ) -> None:
        super().__init__()
        self.database = database
        self.dry_run = dry_run
        self.keep_model = keep_model
        self.wide_outputs = wide_outputs
        self.compiler = Compiler()
        self.connection = connection
        self.config = config or Config()
        self.intrinsic_overrides = intrinsic_overrides
        self._resources = None
        self._last_model = None
        self._last_sources_version = (-1, None)

    @property
    def resources(self):
        if not self._resources:
            with debugging.span("create_session"):
                self.dry_run |= bool(self.config.get("compiler.dry_run", False))
                # NOTE: language="lqp" is not strictly required for LQP execution, but it
                # will significantly improve performance.
                self._resources = create_resources_instance(
                    config=self.config,
                    dry_run=self.dry_run,
                    connection=self.connection,
                    language="lqp",
                )
                if not self.dry_run:
                    self.engine = self._resources.get_default_engine_name()
                    if not self.keep_model:
                        atexit.register(self._resources.delete_graph, self.database, True, "lqp")
        return self._resources

    # Checks the graph index and updates it if necessary
    def prepare_data(self):
        # Has to happen first, so self.dry_run is populated.
        resources = self.resources

        if self.dry_run:
            return

        from relationalai.semantics.snowflake import Table
        table_sources = Table._used_sources
        if not table_sources.has_changed(self._last_sources_version):
            return

        model = self.database
        app_name = resources.get_app_name()
        engine_name = self.engine
        engine_size = self.resources.config.get_default_engine_size()

        program_span_id = debugging.get_program_span_id()
        sources = [t._fqn for t in Table._used_sources]
        self._last_sources_version = Table._used_sources.version()

        assert self.engine is not None

        with debugging.span("poll_use_index", sources=sources, model=model, engine=engine_name):
            resources.maybe_poll_use_index(
                app_name=app_name,
                sources=sources,
                model=model,
                engine_name=self.engine,
                engine_size=engine_size,
                program_span_id=program_span_id,
            )

    def report_errors(self, problems: list[dict[str, Any]], abort_on_error=True):
        from relationalai import errors
        all_errors = []
        undefineds = []
        pyrel_errors = defaultdict(list)
        pyrel_warnings = defaultdict(list)

        for problem in problems:
            message = problem.get("message", "")
            report = problem.get("report", "")
            # TODO: we need to build source maps
            # path = problem.get("path", "")
            # source_task = self._install_batch.line_to_task(path, problem["start_line"]) or task
            # source = debugging.get_source(source_task) or debugging.SourceInfo()
            source = debugging.SourceInfo()
            severity = problem.get("severity", "warning")
            code = problem.get("code")

            if severity in ["error", "exception"]:
                if code == "UNDEFINED_IDENTIFIER":
                    match = re.search(r'`(.+?)` is undefined', message)
                    if match:
                        undefineds.append((match.group(1), source))
                    else:
                        all_errors.append(errors.RelQueryError(problem, source))
                elif "overflowed" in report:
                    all_errors.append(errors.NumericOverflow(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_errors[problem["props"]["pyrel_id"]].append(problem)
                elif abort_on_error:
                    all_errors.append(errors.RelQueryError(problem, source))
            else:
                if code == "ARITY_MISMATCH":
                    errors.ArityMismatch(problem, source)
                elif code == "IC_VIOLATION":
                    all_errors.append(errors.IntegrityConstraintViolation(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_warnings[problem["props"]["pyrel_id"]].append(problem)
                else:
                    errors.RelQueryWarning(problem, source)

        if abort_on_error and len(undefineds):
            all_errors.append(errors.UninitializedPropertyException(undefineds))

        if abort_on_error:
            for pyrel_id, pyrel_problems in pyrel_errors.items():
                all_errors.append(errors.ModelError(pyrel_problems))

        for pyrel_id, pyrel_problems in pyrel_warnings.items():
            errors.ModelWarning(pyrel_problems)


        if len(all_errors) == 1:
            raise all_errors[0]
        elif len(all_errors) > 1:
            raise errors.RAIExceptionSet(all_errors)

    def _export(self, txn_id: str, export_info: tuple, dest: Table, actual_cols: list[str], declared_cols: list[str], update: bool):
        # At this point of the export, we assume that a CSV file has already been written
        # to the Snowflake Native App stage area. Thus, the purpose of this method is to
        # copy the data from the CSV file to the destination table.
        _exec = self.resources._exec
        dest_database, dest_schema, dest_table, _ = IdentityParser(dest._fqn, require_all_parts=True).to_list()
        filename = export_info[0]
        result_table_name = filename + "_table"

        with debugging.span("export", txn_id=txn_id, export_info=export_info, dest_table=dest_table):
            with debugging.span("export_to_result_schema"):
                # First, we need to persist from the CSV file to the results schema by calling the
                # `persist_from_stage` stored procedure. This step also cleans up the CSV file in
                # the stage area.
                column_fields = []
                for (col_name, col_type) in export_info[1]:
                    column_fields.append([col_name, lqp_type_to_sql(col_type)])

                # NOTE: the `str(column_fields)` depends on python formatting which surrounds
                # strings with single quotes. If this changes, or if we ever get a single quote in
                # the actual string, then we need to do something more sophisticated.
                exec_str = f"call {APP_NAME}.api.persist_from_stage('{txn_id}', '{filename}', '{result_table_name}', {str(column_fields)})"
                _exec(exec_str)

            with debugging.span("write_table"):
                # The result of the first step above is a table in the results schema,
                # {app_name}.results.{result_table_name}.
                # Second, we need to copy the data from the results schema to the
                # destination table. This step also cleans up the result table.
                out_sample = _exec(f"select * from {APP_NAME}.results.{result_table_name} limit 1;")
                names = self._build_projection(declared_cols, actual_cols, column_fields, out_sample)
                dest_fqn = dest._fqn
                try:
                    if not update:
                        createTableLogic = f"""
                                        CREATE TABLE {dest_fqn} AS
                                        SELECT {names}
                                        FROM {APP_NAME}.results.{result_table_name};
                        """
                        if dest._is_iceberg:
                            assert dest._iceberg_config is not None
                            external_volume_clause = ""
                            if dest._iceberg_config.external_volume:
                                external_volume_clause = f"EXTERNAL_VOLUME = '{dest._iceberg_config.external_volume}'"
                            createTableLogic = f"""
                                        CREATE ICEBERG TABLE {dest_fqn}
                                        CATALOG = "SNOWFLAKE"
                                        {external_volume_clause}
                                        AS
                                        SELECT {names}
                                        FROM {APP_NAME}.results.{result_table_name};
                            """

                        _exec(f"""
                            BEGIN
                                -- Check if table exists
                                IF (EXISTS (
                                    SELECT 1
                                    FROM {dest_database}.INFORMATION_SCHEMA.TABLES
                                    WHERE table_schema = '{dest_schema}'
                                    AND table_name = '{dest_table}'
                                )) THEN
                                    -- Insert into existing table
                                    EXECUTE IMMEDIATE '
                                        BEGIN
                                            TRUNCATE TABLE {dest_fqn};
                                            INSERT INTO {dest_fqn}
                                            SELECT {names}
                                            FROM {APP_NAME}.results.{result_table_name}
                                            {'' if out_sample else 'WHERE 1=0'};
                                        END;
                                    ';
                                ELSE
                                    -- Create table based on the SELECT
                                    EXECUTE IMMEDIATE '
                                        {escape_for_f_string(createTableLogic)}
                                    ';
                                END IF;
                            END;
                        """)
                    else:
                        if out_sample:
                            _exec(f"""
                                BEGIN
                                    INSERT INTO {dest_fqn}
                                    SELECT {names}
                                    FROM {APP_NAME}.results.{result_table_name};
                                END;
                            """)
                finally:
                    # Always try to drop the result table, even if the insert/create failed.
                    _exec(f"call {APP_NAME}.api.drop_result_table('{result_table_name}');")

    def _build_projection(self, declared_cols, actual_cols, column_fields, out_sample=None):
        # map physical col -> type
        col_type_map = {col.lower(): dtype for col, dtype in column_fields}
        sample_keys = {k.lower() for k in out_sample[0].as_dict()} if out_sample else set()

        fields = []
        ix = 0

        for name in declared_cols:
            if name not in actual_cols:
                # Declared but not present in results
                fields.append(f"NULL as \"{name}\"")
                continue

            # Get the actual physical column name from column_fields
            colname = column_fields[ix][0]
            ix += 1

            if colname.lower() in sample_keys:
                # Actual column exists in sample
                fields.append(f"{colname} as \"{name}\"")
            else:
                # No sample or missing key → fall back to type
                dtype = col_type_map.get(colname.lower(), "VARCHAR")
                fields.append(f"CAST(NULL AS {dtype}) as \"{name}\"")

        return ", ".join(fields)

    def _construct_configure(self):
        config_dict = {}
        # Only set the IVM flag if there is a value in `config`. Otherwise, let
        # `construct_configure` set the default value.
        ivm_flag = self.config.get('reasoner.rule.incremental_maintenance', None)
        if ivm_flag:
            config_dict['ivm.maintenance_level'] = lqp_bundle.ir.Value(value=ivm_flag, meta=None)

        # Set semantics_version from config, defaulting to 0
        semantics_version: str | Any = self.config.get('reasoner.rule.lqp.semantics_version', DEFAULT_LQP_SEMANTICS_VERSION)
        config_dict['semantics_version'] = lqp_bundle.ir.Value(value=int(semantics_version), meta=None)

        # Warn if a non-default semantics version is used. Most likely, this is due to a
        # user manually reverting to an older version. We want them to not get stuck on that
        # version for longer than necessary.
        if semantics_version != DEFAULT_LQP_SEMANTICS_VERSION:
            debugging.warn(NonDefaultLQPSemanticsVersionWarning(semantics_version, DEFAULT_LQP_SEMANTICS_VERSION))

        return lqp_bundle.construct_configure(config_dict, None)

    def _should_sync(self, model) :
        if self._last_model != model:
            return lqp_bundle.ir.Sync(fragments=[], meta=None)
        else :
            return None

    def _compile_intrinsics(self) -> lqp_bundle.ir.Epoch:
        """Construct an epoch that defines a number of built-in definitions used by the
        emitter."""
        with debugging.span("compile_intrinsics") as span:
            span["compile_type"] = "intrinsics"

            now = self.intrinsic_overrides.get('datetime_now', datetime.now(timezone.utc))

            debug_info = lqp_bundle.ir.DebugInfo(id_to_orig_name={}, meta=None)
            intrinsics_fragment = lqp_bundle.ir.Fragment(
                id = lqp_bundle.ir.FragmentId(id=b"__pyrel_lqp_intrinsics", meta=None),
                declarations = [mk_intrinsic_datetime_now(now)],
                debug_info = debug_info,
                meta = None,
            )

            return lqp_bundle.ir.Epoch(
                writes=[
                    lqp_bundle.ir.Write(write_type=lqp_bundle.ir.Define(fragment=intrinsics_fragment, meta=None), meta=None)
                ],
                meta=None,
            )

    def _add_nomaintain(self, epoch: lqp_bundle.ir.Epoch) -> lqp_bundle.ir.Epoch:
        """Add the nomaintain attribute to all Def nodes in an LQP epoch."""
        _nomaintain = lqp_bundle.ir.Attribute(name="nomaintain", args=[], meta=None)
        new_writes = []
        for write in epoch.writes:
            if not isinstance(write.write_type, lqp_bundle.ir.Define):
                new_writes.append(write)
                continue
            fragment = write.write_type.fragment
            new_decls = [
                replace(decl, attrs=list(decl.attrs) + [_nomaintain])
                if isinstance(decl, lqp_bundle.ir.Def) and not any(a.name == "nomaintain" for a in decl.attrs)
                else decl
                for decl in fragment.declarations
            ]
            new_fragment = replace(fragment, declarations=new_decls)
            new_define = replace(write.write_type, fragment=new_fragment)
            new_writes.append(replace(write, write_type=new_define))
        return replace(epoch, writes=new_writes)

    # [RAI-40997] We eagerly undefine query fragments so they are not committed to storage
    def _compile_undefine_query(self, query_epoch: lqp_bundle.ir.Epoch) -> lqp_bundle.ir.Epoch:
        fragment_ids = []

        for write in query_epoch.writes:
            if isinstance(write.write_type, lqp_bundle.ir.Define):
                fragment_ids.append(write.write_type.fragment.id)

        # Construct new Epoch with Undefine operations for all collected fragment IDs
        undefine_writes = [
            lqp_bundle.ir.Write(
                write_type=lqp_bundle.ir.Undefine(fragment_id=frag_id, meta=None),
                meta=None
            )
            for frag_id in fragment_ids
        ]

        return lqp_bundle.ir.Epoch(
            writes=undefine_writes,
            meta=None,
        )

    def _extract_export_uri_short_name(self, config_path: str) -> str:
        # config.path is a full Snowflake stage URI of the form:
        #   snowflake://<stage>/export/<short_name>/data
        # export_info stores the short_name (e.g. "export_<uuid>").
        # We parse the URI to extract the short_name for comparison
        try:
            parsed = urlparse(config_path)
            path_parts = parsed.path.strip("/").split("/")
        except Exception as e:
            raise ValueError(f"Failed to parse export URI {config_path!r}: {e}")
        # Expected structure: ["export", "<short_name>", "data"]
        if not (len(path_parts) == 3 and path_parts[0] == "export" and path_parts[2] == "data"):
            raise ValueError(
                f"Unexpected export URI path structure: {config_path!r}. "
                f"Expected 'snowflake://<stage>/export/<name>/data'."
            )
        return path_parts[1]

    def _filter_csv_exports(self, task: ir.Task, query_epoch: lqp_bundle.ir.Epoch, export_info: Optional[ExportInfo]) -> lqp_bundle.ir.Epoch:
        """
        Filter CSV export columns based on the original column names from Output nodes.

        Maps export reads to their corresponding outputs by FQN and filters out extra columns
        that were added during processing but weren't in the original output specification.
        """

        if not export_info:
            return query_epoch

        # Count user-specified columns per export from Output nodes
        outputs = collect_by_type(ir.Output, task)
        assert outputs, "No Output found in the task"
        n_user_cols: dict[str, int] = {}
        for output in outputs:
            export_fqn = extract_export_fqn(output.annotations)
            if export_fqn:
                n_user_cols[export_fqn] = sum(1 for alias, _ in output.aliases if alias)

        # Validate that exports in outputs and export_info are consistent
        export_fqns_from_info = {fqn for fqn, _, _ in export_info}
        missing_in_info = set(n_user_cols.keys()) - export_fqns_from_info
        if missing_in_info:
            raise ValueError(f"Export FQNs found in outputs but missing in export_info: {missing_in_info}")
        extra_in_info = export_fqns_from_info - set(n_user_cols.keys())
        if extra_in_info:
            raise ValueError(f"Export FQNs found in export_info but missing in outputs: {extra_in_info}")

        # Build mapping: export_fqn -> list of column names from export_info.
        # We use export_info names (not Output.aliases) because the compiler may
        # rename Hash-typed columns (e.g. NODE -> NODE_0_uuid).
        exports_cols: dict[str, list[str]] = {}
        for fqn, _, cols in export_info:
            n = n_user_cols[fqn]
            # The compiler preserves user-specified column order and places them first in export_info
            # cols entries are either (col_name, col_type) tuples or plain column name strings
            exports_cols[fqn] = [c[0] if isinstance(c, tuple) else c for c in cols[:n]]

        # Apply filter to each export read by matching FQN
        if query_epoch.reads:
            new_reads = []
            for read in query_epoch.reads:
                if isinstance(read.read_type, lqp_bundle.ir.Export):
                    # For each Export read, extract the short name from its config path
                    # and find the matching export FQN
                    config_path = read.read_type.config.path
                    short_name = self._extract_export_uri_short_name(config_path)

                    matching_fqn = None
                    if export_info:
                        for fqn, filename, _ in export_info:
                            if filename == short_name:
                                matching_fqn = fqn
                                break

                    if matching_fqn and matching_fqn in exports_cols:
                        # Apply filter for this export using its FQN
                        cols = exports_cols[matching_fqn]
                        column_filter = export_rewriter.ExtraColumnsFilter(cols)
                        read = column_filter.filter_read(read)
                    # Disable partitioning for all Export reads in CSV format
                    read = export_rewriter.PartitioningDisabler().rewrite_read(read)
                    new_reads.append(read)
                else:
                    new_reads.append(read)

            return replace(query_epoch, reads=new_reads)

        return query_epoch

    def compile_lqp(self, model: ir.Model, task: ir.Task, format: Optional[Literal["pandas", "snowpark", "csv"]] = "pandas"):
        configure = self._construct_configure()
        # Merge the epochs into a single transaction. Long term the query bits should all
        # go into a WhatIf action and the intrinsics could be fused with either of them. But
        # for now we just use separate epochs.
        epochs = []
        epochs.append(self._compile_intrinsics())

        sync = self._should_sync(model)

        if sync is not None:
            with debugging.span("compile", metamodel=model) as install_span:
                install_span["compile_type"] = "model"
                _, model_epoch = self.compiler.compile(model, {"fragment_id": b"model", "_v0_config": self.config})
                epochs.append(model_epoch)
                self._last_model = model

        with debugging.span("compile", metamodel=task) as txn_span:
            query = f.compute_model(f.logical([task]))
            options = {
                "wide_outputs": self.wide_outputs,
                "fragment_id": b"query",
                "_v0_config": self.config,
            }
            result, final_model = self.compiler.compile_inner(query, options)
            export_info, query_epoch = result

            query_epoch = self._add_nomaintain(query_epoch)

            if format == "csv":
                # Filter CSV exports to only include original columns
                query_epoch = self._filter_csv_exports(task, query_epoch, export_info)

            # Adds constraints declared in previous epochs to the query epoch. A temporary
            # workaround for limited FD Salsa storage in the backend See: https://github.com/RelationalAI/relationalai-python/pull/3262
            _add_previous_constraint_declarations(query_epoch, epochs)

            epochs.append(query_epoch)
            epochs.append(self._compile_undefine_query(query_epoch))

            txn_span["compile_type"] = "query"
            txn = mk_transaction(epochs=epochs, configure=configure, sync=sync)
            txn_span["lqp"] = lqp_bundle.to_safe_string(txn)

        txn_proto = lqp_bundle.ir_to_proto(txn)
        validate_proto(txn_proto)
        return final_model, export_info, txn_proto

    def execute(
        self,
        model: ir.Model,
        task: ir.Task,
        format: Literal["pandas", "snowpark", "csv"] = "pandas",
        export_to: Optional[Table] = None,
        update: bool = False,
        meta: dict[str, Any] | None = None,
        source_info: dict[str, Any] | None = None,
    ) -> DataFrame:
        self.prepare_data()
        previous_model = self._last_model
        final_model, export_info, txn_proto = self.compile_lqp(model, task, format=format)

        if self.dry_run:
            return DataFrame()

        if format == "snowpark":
            raise ValueError(f"Unsupported format: {format}")

        # Format meta as headers
        json_meta = prepare_metadata_for_headers(meta)
        headers = {QUERY_ATTRIBUTES_HEADER: json_meta} if json_meta else {}
        raw_results = self.resources.exec_lqp(
            self.database,
            self.engine,
            txn_proto.SerializeToString(),
            # Current strategy is to run all queries as write transactions, in order to
            # benefit from view caching. This will have to be revisited, because write
            # transactions are serialized.
            readonly=False,
            nowait_durable=True,
            headers=headers,
            source_info=source_info,
        )
        assert isinstance(raw_results, TransactionAsyncResponse), "Expected TransactionAsyncResponse from LQP execution"
        assert raw_results.transaction is not None, "Transaction result is missing"
        txid = raw_results.transaction['id']

        try:
            # For CSV format, return paths directly without computing columns
            if format == "csv":
                if export_info and isinstance(export_info, list):
                    # export_info is a list of (fqn, filename, col_types)
                    paths = []
                    for export_entry in export_info:
                        filename = export_entry[1]  # (fqn, filename, col_types) -> filename
                        full_path = filename + f"/data_{txid}.gz"
                        paths.append(full_path)
                    return DataFrame(paths, columns=["path"])
                else:
                    raise ValueError("The CSV export was not successful!")

            cols, extra_cols, key_locs = self._compute_cols(task, final_model)
            df, errs = result_helpers.format_results(raw_results, cols, key_locs)
            self.report_errors(errs)

            if export_to:
                # Single export with explicit destination table. Multiple table exports are not supported yet.
                assert cols, "No columns found in the output"
                assert export_info, "Export info should be populated if we are exporting results"
                assert isinstance(export_info, list), "export_info must be a list"
                assert len(export_info) == 1, "export_to only supports single export"
                result_cols = export_to._col_names
                if result_cols is not None:
                    assert all(col in result_cols or col in extra_cols for col in cols)
                else:
                    result_cols = [col for col in cols if col not in extra_cols]
                assert result_cols
                # Extract single export info: (fqn, filename, col_types)
                single_export_info = (export_info[0][1], export_info[0][2])
                self._export(txid, single_export_info, export_to, cols, result_cols, update)

            return self._postprocess_df(self.config, df, extra_cols)

        except Exception as e:
            # If processing the results failed, revert to the previous model.
            self._last_model = previous_model
            raise e

    def export_to_csv(self, model: "InternalModel", query) -> str:
        ### Exports the result of the given query fragment to a CSV file in
        ### the Snowflake stage area and returns the path to the CSV file.

        from relationalai.semantics.internal.internal import Fragment, with_source
        from relationalai.environments import runtime_env

        if not query._select:
            raise ValueError("Cannot export empty selection to CSV")

        clone = Fragment(parent=query)
        clone._is_export = True
        clone._source = runtime_env.get_source_pos()
        ir_model = model._to_ir()
        source_info = with_source(clone)
        with debugging.span("query", dsl=str(clone), **source_info, meta=clone._meta):
            query_task = model._compiler.fragment(clone)
            csv_info = self.execute(ir_model, query_task, format="csv", meta=clone._meta, source_info=source_info)
            if csv_info.empty or "path" not in csv_info.columns:
                return ""
            else:
                path = csv_info.at[0, "path"]
                assert isinstance(path, str)
            return path

    def batch_export_to_csv(self, model: "InternalModel", queries: list) -> list[str]:
        """Export multiple query fragments to CSV in a single LQP transaction.

        This is an internal method used by the optimization library where batching
        exports into a single transaction is needed for adequate performance. It is
        not meant to be relied on by anyone else and may change without notice.

        All fragments are compiled together as a single combined IR task. Each
        fragment gets a unique export FQN annotation so the compiler treats them
        as independent exports within the same transaction. This reduces round
        trips from N to 1.

        Parameters
        ----------
        queries : list[Fragment]
            List of query fragments to export. Each must contain a SELECT.

        Returns
        -------
        list[str]
            List of CSV file paths, one per input query, in the same order.
        """
        from relationalai.semantics.internal.internal import Fragment
        from relationalai.semantics.metamodel import builtins, types as mm_types
        from relationalai.environments import runtime_env

        if not queries:
            return []

        # Clone each query, mark as export with a unique FQN annotation, compile to IR task.
        # The unique FQN ensures the UnifyDefinitions rewrite pass does not merge
        # different exports, and the LQP compiler creates separate ExportCSVConfigs.
        tasks = []
        for i, query in enumerate(queries):
            if not query._select:
                raise ValueError("Cannot export empty selection to CSV")
            clone = Fragment(parent=query)
            clone._is_export = True
            clone._source = runtime_env.get_source_pos()
            clone._annotations = [f.annotation(builtins.export, [ir.Literal(mm_types.String, f"__batch_export_{i}")])]
            tasks.append(model._compiler.fragment(clone))

        # Combine all tasks into a single logical task
        combined_task = f.logical(tasks)
        ir_model = model._to_ir()

        meta = {"operation": "batch_export_to_csv", "num_exports": len(queries)}
        with debugging.span("batch_export_to_csv", num_exports=len(queries)):
            csv_info = self.execute(ir_model, combined_task, format="csv", meta=meta)

        if csv_info.empty or "path" not in csv_info.columns:
            raise ValueError("Batch CSV export returned no paths")

        paths = csv_info["path"].tolist()
        assert len(paths) == len(queries), (
            f"Expected {len(queries)} CSV paths but got {len(paths)}"
        )
        return paths

# The following two functions are a temporary workaround for limited FD Salsa storage in the
# backend See: https://github.com/RelationalAI/relationalai-python/pull/3262

# Adds constraints declared in previous epochs to the query epoch.
def _add_previous_constraint_declarations(query_epoch: lqp_bundle.ir.Epoch, prev_epochs: list[lqp_bundle.ir.Epoch]):
    constraints: list[lqp_bundle.ir.Constraint] = []

    # Find all declared constraints in previous epochs
    for epoch in prev_epochs:
        for write in epoch.writes:
            if isinstance(write.write_type, lqp_bundle.ir.Define):
                fragment = write.write_type.fragment
                for decl in fragment.declarations:
                    if isinstance(decl, lqp_bundle.ir.Constraint):
                        constraints.append(decl)

    # Add the collected constraints to the define frags of the query epoch
    for write in query_epoch.writes:
        if isinstance(write.write_type, lqp_bundle.ir.Define):
            fragment = write.write_type.fragment
            # Prepend the collected constraints to the declarations
            assert isinstance(fragment.declarations, list)
            fragment.declarations.extend(_fd_with_next_id(fd) for fd in constraints)

# Creates a copy of the given FD with a new ID to avoid multiple declarations with the same ID.
def _fd_with_next_id(fd: lqp_bundle.ir.Constraint) -> lqp_bundle.ir.Constraint:
    assert isinstance(fd, lqp_bundle.ir.FunctionalDependency)
    next_id = utils.lqp_hash(str((fd.name,)))
    lqp_name = lqp_bundle.ir.RelationId(id=next_id, meta=None)

    return lqp_bundle.ir.FunctionalDependency(
        name=lqp_name,
        guard=fd.guard,
        keys=fd.keys,
        values=fd.values,
        meta=None
    )



from __future__ import annotations
from dataclasses import dataclass, field
from relationalai.clients.config import Config
from relationalai.semantics.metamodel import builtins
from relationalai.semantics.metamodel.ir import Node, Model, Require
from relationalai.semantics.metamodel.compiler import Pass
from relationalai.semantics.metamodel.rewrite.discharge_constraints import (
    DischargeConstraintsVisitor
)
from relationalai.semantics.lqp.rewrite.functional_dependencies import (
    normalized_fds
)

class AnnotateConstraints(Pass):
    """
    Extends `DischargeConstraints` pass by discharging only those Require nodes that cannot
    be declared as constraints in LQP.

    More precisely, the pass annotates Require nodes depending on how they should be
    treated when generating code:
     * `@declare_constraint` if the Require represents a constraint that can be declared in LQP.
     * `@discharge` if the Require represents a constraint that should be dismissed during
       code generation. Namely, when it cannot be declared in LQP and uses one of the
       `unique`, `exclusive`, `anyof` builtins. These nodes are removed from the IR model
       in the Flatten pass.
    """

    def rewrite(self, model: Model, options: dict = {}) -> Model:
        # Extract config from options if provided (for compatibility with PyRel ShimExecutor)
        config = options.get('_v0_config', None)
        return AnnotateConstraintsRewriter(config=config).walk(model)

@dataclass
class AnnotateConstraintsRewriter(DischargeConstraintsVisitor):
    emit_constraints: bool = field(default=False)
    config: Config | None = field(default=None, repr=False)
    """
    Visitor marks Require nodes with
    - `discharge` if they should be discharged from the metamodel
    - `declare_constraint` if they should be kept and emitted as LQP constraint declarations

    By default, all constraints are discharged. To enable emitting constraints, set the
    `reasoner.rule.emit_constraints` flag to True in the config file.
    ```toml
    [reasoner.rule]
    emit_constraints = true
    ```
    """

    def __post_init__(self):
        from relationalai.semantics.internal.internal import overridable_flag
        # Use provided config if available, otherwise try to create one with fetch=False
        # to avoid requiring raiconfig.toml when config is passed programmatically
        if self.config is None:
            try:
                self.config = Config(fetch=False)
            except Exception:
                # If Config() fails even with fetch=False, use empty dict
                self.config = Config({}, fetch=False)
        self.emit_constraints = overridable_flag('reasoner.rule.emit_constraints', self.config, None, False)

    def _should_declare_constraint(self, node: Require) -> bool:
        if not self.emit_constraints:
            return False
        fds = normalized_fds(node)
        if len(fds) == 0:
            return False
        # Currently, we only declare non-structural functional dependencies.
        return any(not fd.is_structural for fd in fds)

    def handle_require(self, node: Require, parent: Node):
        if self._should_declare_constraint(node):
            return node.reconstruct(
                node.engine,
                node.domain,
                node.checks,
                node.annotations | [builtins.declare_constraint_annotation]
            )

        return super().handle_require(node, parent)
